import pytest
from wait_for import wait_for
from widgetastic.widget import ParametrizedView
from widgetastic.widget import Text

from widgetastic_patternfly4 import CardCheckBox
from widgetastic_patternfly4 import CardForCardGroup
from widgetastic_patternfly4 import CardGroup
from widgetastic_patternfly4 import Dropdown

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/demos/card-view/react-demos/card-view/"


class PageCard(CardForCardGroup):
    dropdown = Dropdown()

    def delete_action(self):
        self.dropdown.item_select("Delete")

    checked = CardCheckBox()

    header_text = Text(locator="./div[@class='pf-c-card__title']")


class Cards(CardGroup):
    def __init__(self, parent, locator=None, logger=None, **kwargs):
        super().__init__(parent, logger=logger, **kwargs)
        self.locator = locator or './/div[contains(@class, "pf-l-gallery")]'

    cards = ParametrizedView.nested(PageCard)


@pytest.fixture
def cards(browser):

    cards = Cards(browser)
    wait_for(lambda: cards.is_displayed, timeout="15s")
    return cards


def test_read_and_drop_second_card(cards):
    second = list(cards)[1]

    assert second.header_text.read() == "Patternfly"

    second.delete_action()

    new_second = list(cards)[1]

    assert new_second.header_text.read() != "Patternfly"


def read_cards_2_checkmap(cards):
    data = cards.cards.read()
    return {card["header_text"]: card["checked"] for card in list(data.values())[1:]}


def test_select_all_cards(browser, cards):
    name2checked = read_cards_2_checkmap(cards)
    assert not any(name2checked.values())
    assert all(name2checked.keys())

    # first card doesn't have header and checkbox
    for card in list(cards)[1:]:
        browser.execute_script("arguments[0].scrollIntoView({block: 'center'});", card.checked)
        card.checked.fill(True)

    name2checked_after = read_cards_2_checkmap(cards)
    assert all(name2checked_after.values())
    assert all(name2checked_after.keys())

    assert name2checked.keys() == name2checked_after.keys()



import pytest
from widgetastic.widget import Checkbox
from widgetastic.widget import View

from widgetastic_patternfly4 import ColumnNotExpandable
from widgetastic_patternfly4 import CompoundExpandableTable
from widgetastic_patternfly4 import ExpandableTable
from widgetastic_patternfly4 import PatternflyTable
from widgetastic_patternfly4 import RowNotExpandable

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/table"

SORT = [
    (
        "This is a really long table header that goes on for a long time 1.",
        "ascending",
        ["a", "one", "p"],
    ),
    (
        "This is a really long table header that goes on for a long time 1.",
        "descending",
        ["p", "one", "a"],
    ),
    (
        "This is a really long table header that goes on for a long time 3.",
        "ascending",
        ["a", "b", "k"],
    ),
    (
        "This is a really long table header that goes on for a long time 3.",
        "descending",
        ["k", "b", "a"],
    ),
]


@pytest.mark.parametrize("sample", SORT, ids=lambda sample: "{}-{}".format(sample[0], sample[1]))
def test_sortable_table(browser, sample):
    header, order, expected_result = sample
    table = PatternflyTable(
        browser,
        ".//div[@id='ws-react-composable-c-table-composable-sortable--wrapping-headers']/table",
    )
    table.sort_by(header, order)
    column = [row[header] for row in table.read()]
    assert column == expected_result


@pytest.mark.parametrize(
    "sample", [("select_all", True), ("deselect_all", False)], ids=("select", "deselect")
)
def test_selectable_table(browser, sample):
    method, expected_result = sample
    table = PatternflyTable(
        browser,
        ".//div[@id='ws-react-composable-c-table-composable-selectable-with-checkbox']//table",
        column_widgets={0: Checkbox(locator=".//input")},
    )
    getattr(table, method)()
    for row in table:
        if row.index != 1:  # skip row with disabled checkbox
            assert expected_result == row[0].widget.selected


def test_expandable_table(browser):
    expected_read = [
        {
            "Repositories": "one",
            "Branches": "two",
            "Pull requests": "a",
            "Workspaces": "four",
            "Last commit": "five",
        },
        {
            "Repositories": "parent 1",
            "Branches": "two",
            "Pull requests": "k",
            "Workspaces": "four",
            "Last commit": "five",
        },
        {
            "Repositories": "parent 2",
            "Branches": "two",
            "Pull requests": "b",
            "Workspaces": "four",
            "Last commit": "five",
        },
        {
            "Repositories": "parent 3",
            "Branches": "2",
            "Pull requests": "b",
            "Workspaces": "four",
            "Last commit": "five",
        },
        {
            "Repositories": "parent 4",
            "Branches": "2",
            "Pull requests": "b",
            "Workspaces": "four",
            "Last commit": "five",
        },
        {
            "Repositories": "parent 5",
            "Branches": "2",
            "Pull requests": "b",
            "Workspaces": "four",
            "Last commit": "five",
        },
        {
            "Repositories": "parent 6",
            "Branches": "2",
            "Pull requests": "b",
            "Workspaces": "four",
            "Last commit": "five",
        },
    ]

    row1_expected_content = "single cell"
    row2_expected_content = "single cell - fullWidth"
    row3_expected_content = "single cell - noPadding"

    table = ExpandableTable(
        browser, ".//div[@id='ws-react-composable-c-table-composable-expandable']/table"
    )

    assert table.read() == expected_read

    # First row is not an expandable row
    assert not table[0].is_expandable
    with pytest.raises(RowNotExpandable):
        table[0].expand()

    parent1_row = table[1]
    parent2_row = table[2]
    parent3_row = table[3]

    parent1_row.collapse()  # The row starts out expanded on the demo page
    assert not parent1_row.is_expanded
    assert not parent1_row.content.is_displayed

    parent1_row.expand()
    assert parent1_row.is_expanded
    assert parent1_row.content.is_displayed
    assert parent1_row.content.read() == row1_expected_content

    parent2_row.expand()
    assert parent2_row.is_expanded
    assert parent2_row.content.is_displayed
    assert parent2_row.content.read() == row2_expected_content

    parent3_row.expand()
    assert parent3_row.is_expanded
    assert parent3_row.content.is_displayed
    assert parent3_row.content.read() == row3_expected_content


@pytest.mark.parametrize(
    "use_different_widgets",
    [True, False],
    ids=["diff-widgets-expandable-content", "same-widget-expandable-content"],
)
def test_compound_expandable_table(browser, use_different_widgets):
    table_read = [
        {
            "Repositories": "siemur/test-space",
            "Branches": "10",
            "Pull requests": "4",
            "Workspaces": "4",
            "Last commit": "20 minutes",
            5: "Open in Github",
        },
        {
            "Repositories": "siemur/test-space",
            "Branches": "3",
            "Pull requests": "4",
            "Workspaces": "2",
            "Last commit": "10 minutes",
            5: "Open in Github",
        },
    ]

    row0_branches_read = {
        "table": [
            {
                "Repositories": "parent-0",
                "Branches": "compound-1",
                "Pull requests": "three",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "a",
                "Branches": "two",
                "Pull requests": "k",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "p",
                "Branches": "two",
                "Pull requests": "b",
                "Workspaces": "four",
                "Last Commit": "five",
            },
        ]
    }
    row0_pull_requests_read = {
        "table": [
            {
                "Repositories": "parent-0",
                "Branches": "compound-2",
                "Pull requests": "three",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "a",
                "Branches": "two",
                "Pull requests": "k",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "p",
                "Branches": "two",
                "Pull requests": "b",
                "Workspaces": "four",
                "Last Commit": "five",
            },
        ]
    }
    row0_workspaces_read = {
        "table": [
            {
                "Repositories": "parent-0",
                "Branches": "compound-3",
                "Pull requests": "three",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "a",
                "Branches": "two",
                "Pull requests": "k",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "p",
                "Branches": "two",
                "Pull requests": "b",
                "Workspaces": "four",
                "Last Commit": "five",
            },
        ]
    }
    row1_branches_read = {
        "table": [
            {
                "Repositories": "parent-1",
                "Branches": "compound-1",
                "Pull requests": "three",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "a",
                "Branches": "two",
                "Pull requests": "k",
                "Workspaces": "four",
                "Last Commit": "five",
            },
            {
                "Repositories": "p",
                "Branches": "two",
                "Pull requests": "b",
                "Workspaces": "four",
                "Last Commit": "five",
            },
        ]
    }
    if use_different_widgets:
        # for the example table all the expanded tables are the same, besides the different id
        # that we use in the locator
        class _Branches(View):
            table = PatternflyTable(locator=".//table[contains(@id, '_1')]")

        class _PullRequests(View):
            table = PatternflyTable(locator=".//table[contains(@id, '_2')]")

        class _Workspaces(View):
            table = PatternflyTable(locator=".//table[contains(@id, '_3')]")

        content_view = {1: _Branches(), 2: _PullRequests(), 3: _Workspaces()}
    else:
        # use the same content_view for all the tables
        class _ContentView(View):
            """View for the nested table(s) in the expandable columns."""

            table = PatternflyTable(locator=".//table[@aria-label='Sortable Table']")

        content_view = _ContentView()

    table = CompoundExpandableTable(
        browser, ".//table[@aria-label='Compound expandable table']", content_view=content_view
    )

    assert table.read() == table_read

    # Make sure that the appropriate columns are expandable
    for row in table.rows():
        assert not row.repositories.is_expandable
        assert not row.last_commit.is_expandable
        assert row.branches.is_expandable
        assert row.pull_requests.is_expandable
        assert row.workspaces.is_expandable

    # first column is not an expandable column
    with pytest.raises(ColumnNotExpandable):
        table[0][0].expand()
    with pytest.raises(ColumnNotExpandable):
        table[1][0].expand()

    # first row
    row0 = table[0]
    row0.branches.expand()
    assert row0.branches.is_expanded
    assert row0.branches.content.read() == row0_branches_read
    row0.branches.collapse()
    assert not row0.branches.is_expanded

    row0.pull_requests.expand()
    assert row0.pull_requests.is_expanded
    assert row0.pull_requests.content.read() == row0_pull_requests_read
    row0.pull_requests.collapse()
    assert not row0.pull_requests.is_expanded

    row0.workspaces.expand()
    assert row0.workspaces.is_expanded
    assert row0.workspaces.content.read() == row0_workspaces_read
    row0.workspaces.collapse()
    assert not row0.workspaces.is_expanded

    # second row, just test the first expandable column
    row1 = table[1]
    row1.branches.expand()
    assert row1.branches.is_expanded
    assert row1.branches.content.read() == row1_branches_read
    row1.branches.collapse()
    assert not row1.branches.is_expanded



import contextlib

import pytest
from wait_for import wait_for
from widgetastic.widget import View

from widgetastic_patternfly4 import CompactPagination
from widgetastic_patternfly4 import Pagination
from widgetastic_patternfly4 import PaginationNavDisabled

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/pagination"


@contextlib.contextmanager
def _paginator(browser, request, reset_elements_per_page=True):
    paginator_cls, kind = request.param

    class TestView(View):
        ROOT = f".//div[@id='ws-react-c-pagination-{kind}']"
        paginator = paginator_cls(locator="./div")

    paginator = TestView(browser).paginator
    wait_for(lambda: paginator.is_displayed, num_sec=10)
    yield paginator
    try:
        paginator.first_page()
    except PaginationNavDisabled:
        # We are already at the first page...
        pass
    if reset_elements_per_page:
        paginator.set_per_page(20)


@pytest.fixture(
    params=[(Pagination, "top"), (CompactPagination, "compact")],
    ids=["Pagination", "CompactPagination"],
)
def paginator(browser, request):
    with _paginator(browser, request) as result:
        yield result


@pytest.fixture(
    params=[
        (Pagination, "one-page"),
        # there is no compact paginator of this type on the demo page
    ],
    ids=["Pagination"],
)
def one_page_paginator(browser, request):
    with _paginator(browser, request) as result:
        yield result


@pytest.fixture(
    params=[
        (Pagination, "no-items"),
        # there is no compact paginator of this type on the demo page
    ],
    ids=["Pagination"],
)
def no_elements_paginator(browser, request):
    with _paginator(browser, request, reset_elements_per_page=False) as result:
        yield result


# Can't parametrize fixtures:
# https://github.com/pytest-dev/pytest/issues/349
# @pytest.mark.parametrize('params', [ { "paginator": one_page_paginator, "expected_pages": 1}])


def _page_iteration_asserts(paginator, expected_pages):
    page_counter = 0

    with paginator.cache_per_page_value():
        for page in paginator:
            page_counter += 1
    assert page_counter == expected_pages


def test_one_page_iteration(one_page_paginator):
    _page_iteration_asserts(paginator=one_page_paginator, expected_pages=1)


def test_no_elements_iteration(no_elements_paginator):
    _page_iteration_asserts(paginator=no_elements_paginator, expected_pages=0)


def test_first_page(paginator):
    paginator.last_page()
    assert paginator.current_page == 27
    paginator.first_page()
    assert paginator.is_first_disabled
    assert paginator.is_previous_disabled
    assert not paginator.is_next_disabled
    assert not paginator.is_last_disabled
    assert paginator.current_page == 1
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (1, 20)
    assert paginator.total_items == 523


def test_previous_page(paginator):
    paginator.next_page()
    assert paginator.current_page == 2
    paginator.previous_page()
    assert not paginator.is_next_disabled
    assert not paginator.is_last_disabled
    assert paginator.current_page == 1
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (1, 20)
    assert paginator.total_items == 523


def test_next_page(paginator):
    paginator.next_page()
    assert not paginator.is_first_disabled
    assert not paginator.is_previous_disabled
    assert not paginator.is_next_disabled
    assert not paginator.is_last_disabled
    assert paginator.current_page == 2
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (21, 40)
    assert paginator.total_items == 523


def test_last_page(paginator):
    paginator.last_page()
    assert not paginator.is_first_disabled
    assert not paginator.is_previous_disabled
    assert paginator.is_next_disabled
    assert paginator.is_last_disabled
    assert paginator.current_page == 27
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (521, 523)
    assert paginator.total_items == 523


def test_per_page_options(paginator):
    assert paginator.per_page_options == [
        "10 per page",
        "20 per page",
        "50 per page",
        "100 per page",
    ]


@pytest.mark.parametrize("items_per_page", [50, 100])
def test_iteration(paginator, items_per_page):
    assert paginator.is_first_disabled
    paginator.set_per_page(items_per_page)
    assert paginator.current_per_page == items_per_page

    # Ensure we're always using an int for the math calculations
    items_per_page_int = int(str(items_per_page).split()[0])

    expected_total_pages = 523 // items_per_page_int + 1
    assert paginator.is_previous_disabled
    with paginator.cache_per_page_value():
        for page in paginator:
            assert paginator.current_page == page
            assert paginator.total_pages == expected_total_pages
            if items_per_page_int * page > paginator.total_items:
                right_number = paginator.total_items
            else:
                right_number = items_per_page_int * page
            assert paginator.displayed_items == (1 + items_per_page_int * (page - 1), right_number)
            assert paginator.total_items == 523
    assert paginator.is_next_disabled
    assert paginator.is_last_disabled


def test_bad_paginator_page_value(paginator):
    with pytest.raises(ValueError):
        paginator.set_per_page(9999999)
    with pytest.raises(ValueError):
        paginator.set_per_page("999999")


def test_custom_page(paginator):
    if isinstance(paginator, CompactPagination):
        pytest.skip("Cannot insert custom page number into CompactPagination")
    disp_items = paginator.displayed_items
    paginator.go_to_page(2)
    assert paginator.current_page == 2
    assert disp_items != paginator.displayed_items



from widgetastic.exceptions import NoSuchElementException
from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import View
from widgetastic.xpath import quote


class ModalItemNotFound(Exception):
    pass


class BaseModal:
    """Represents the Patternfly Modal.

    https://www.patternfly.org/v4/documentation/react/components/modal
    """

    BODY = ".//div[contains(@class, 'pf-c-modal-box__body')]"
    FOOTER = ".//*[contains(@class, 'pf-c-modal-box__footer')]/child::node()"
    FOOTER_ITEM = (
        ".//*[contains(@class, 'pf-c-modal-box__footer')]" "/button[normalize-space(.)={}]"
    )
    TITLE = ".//h1[contains(@class, 'pf-c-title') or contains(@class, 'pf-c-modal-box__title')]"
    CLOSE = ".//button[@aria-label='Close']"

    @property
    def title(self):
        """Get title of modal window."""
        return self.browser.text(self.browser.element(self.TITLE))

    @property
    def body(self):
        """Get WebElement of modal body."""
        return self.browser.element(self.BODY)

    def close(self):
        """Close modal window."""
        self.browser.click(self.browser.element(self.CLOSE))

    @property
    def footer_items(self):
        """Returns a list of strings of all items in modal footer."""
        return [self.browser.text(el) for el in self.browser.elements(self.FOOTER)]

    def footer_item(self, item):
        """Returns a WebElement for given item name in modal footer."""
        try:
            return self.browser.element(self.FOOTER_ITEM.format(quote(item)))
        except NoSuchElementException:
            raise ModalItemNotFound(f"Item {item} not found. Available items: {self.footer_items}")

    def __repr__(self):
        return f"{type(self).__name__}({self.locator!r})"


class Modal(BaseModal, View):
    ROOT = ParametrizedLocator("{@locator}")

    def __init__(self, parent, locator=None, logger=None, **kwargs):
        super().__init__(parent, logger=logger, **kwargs)
        self.locator = locator or ".//div[contains(@class, 'pf-c-modal-box')]"



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import DropdownItemDisabled
from widgetastic_patternfly4 import DropdownItemNotFound
from widgetastic_patternfly4.ouia import Dropdown

TESTING_PAGE_URL = (
    "https://patternfly-docs-ouia.netlify.app/documentation/react/components/dropdown"  # noqa
)


@pytest.fixture
def dropdown(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-dropdown-ouia']"
        dropdown = Dropdown("Dropdown")

    view = TestView(browser)
    return view.dropdown


def test_dropdown_is_displayed(dropdown):
    assert dropdown.is_displayed


def test_enabled_dropdown(dropdown):
    assert dropdown.is_enabled


def test_dropdown_items(dropdown):
    assert dropdown.items == [
        "Link",
        "Action",
        "Disabled Link",
        "Disabled Action",
        "",
        "Separated Link",
        "Separated Action",
    ]
    assert dropdown.has_item("Action")
    assert not dropdown.has_item("Non existing items")
    assert dropdown.item_enabled("Action")
    assert not dropdown.item_enabled("Disabled Link")


def test_dropdown_open(dropdown):
    assert not dropdown.is_open
    dropdown.open()
    assert dropdown.is_open
    dropdown.close()
    assert not dropdown.is_open


def test_dropdown_item_select(dropdown):
    dropdown.item_select("Action")
    assert not dropdown.is_open
    with pytest.raises(DropdownItemDisabled):
        dropdown.item_select("Disabled Link")
    with pytest.raises(DropdownItemNotFound):
        dropdown.item_select("Non existing items")



from time import sleep

import pytest

from widgetastic_patternfly4 import PieChart
from widgetastic_patternfly4.bulletchart import DataPoint
from widgetastic_patternfly4.bulletchart import Legend

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/charts/pie-chart"

DATA = {"Cats": 35, "Dogs": 55, "Birds": 10}
LEGENDS = [Legend(label, value) for label, value in DATA.items()]
DATA_POINTS = [DataPoint(label, value) for label, value in DATA.items()]


@pytest.fixture(
    params=[
        {
            "id": "ws-react-c-pie-chart-multi-color-ordered-with-bottom-aligned-legend",
            "anchor": "#basic-with-right-aligned-legend",
        },
        {
            "id": "ws-react-c-pie-chart-basic-with-right-aligned-legend",
            "anchor": "#orange-with-right-aligned-legend",
        },
    ],
    ids=["bottom-aligned", "right-aligned"],
)
def chart(browser, request):
    sleep(3)  # Stabilized graph data on testing page; specially for firefox.
    # Firefox fails the test if the chart is not fully visible therefore we click here on anchor
    # in order to properly scroll down
    anchor = browser.element(f".//a[@href='{request.param['anchor']}']")
    browser.click(anchor)
    return PieChart(browser, id=request.param["id"])


def test_pie_chart(chart):
    """Test PieChart widget."""
    assert chart.is_displayed
    assert chart.legends == LEGENDS
    for i in range(3):
        if chart.data == DATA_POINTS:
            break
        if i == 2:
            assert chart.data == DATA_POINTS

    assert "Birds" in chart.legend_names

    # get data point and check values
    birds_legend = chart.get_legend("Birds")
    birds_data = chart.get_data_point("Birds")

    assert birds_data.label == birds_legend.label
    assert birds_data.value == birds_legend.value
    assert birds_data.color == birds_legend.color

    # read graph
    assert chart.read() == DATA



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Popover
from widgetastic_patternfly4.button import Button

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/popover"
PF4_EXAMPLE_POPOVER_TEXT_TITLE = "Popover header"
PF4_EXAMPLE_POPOVER_TEXT_BODY = "Popovers are triggered by click rather than hover."
PF4_EXAMPLE_POPOVER_TEXT_FOOTER = "Popover footer"


@pytest.fixture()
def popover(browser):
    class TestView(View):
        ROOT = ".//body"
        popover = Popover()
        open_button = Button("Toggle popover")

    view = TestView(browser)
    view.open_button.click()
    yield view.popover
    if view.popover.is_displayed:
        view.popover.close()


def test_popover_is_displayed(popover):
    assert popover.is_displayed


def test_popover_title_text_accessible(popover):
    assert popover.title == PF4_EXAMPLE_POPOVER_TEXT_TITLE


def test_popover_body_text_accessible(popover):
    assert popover.body == PF4_EXAMPLE_POPOVER_TEXT_BODY


def test_popover_footer_text_accessible(popover):
    assert popover.footer == PF4_EXAMPLE_POPOVER_TEXT_FOOTER


def test_popover_close(popover):
    popover.close()
    assert not popover.is_displayed



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import CheckboxSelect
from widgetastic_patternfly4 import Select
from widgetastic_patternfly4 import SelectItemNotFound

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/select"


@pytest.fixture
def select(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-select-single']"
        select = Select(locator=".//div[contains(@class, 'pf-c-select')]")

    return TestView(browser).select


def test_select_is_displayed(select):
    assert select.is_displayed


def test_select_items(select):
    assert set(select.items) == {"Select a title", "Mr", "Miss", "Mrs", "Ms", "Dr", "Other"}
    assert select.has_item("Mr")
    assert not select.has_item("Non existing item")
    assert select.item_enabled("Miss")


def test_select_open(select):
    assert not select.is_open
    select.open()
    assert select.is_open
    select.close()
    assert not select.is_open


def test_select_item_select(select):
    select.fill("Mr")
    assert select.read() == "Mr"
    assert not select.is_open
    with pytest.raises(SelectItemNotFound):
        select.fill("Non existing item")
    assert not select.is_open


@pytest.fixture
def checkbox_select(browser):
    class TestView(View):
        ROOT = './/div[@id="ws-react-c-select-checkbox-input"]'
        checkbox_select = CheckboxSelect(locator=".//div[contains(@class, 'pf-c-select')]")

    return TestView(browser).checkbox_select


def test_checkbox_select_is_displayed(checkbox_select):
    assert checkbox_select.is_displayed


def test_checkbox_select_items(checkbox_select):
    assert set(checkbox_select.items) == {
        "Active This is a description",
        "Cancelled",
        "Paused",
        "Warning",
        "Restarted",
    }
    assert checkbox_select.has_item("Active This is a description")
    assert not checkbox_select.has_item("Non existing item")
    assert checkbox_select.item_enabled("Paused")


def test_checkbox_select_open(checkbox_select):
    assert not checkbox_select.is_open
    checkbox_select.open()
    assert checkbox_select.is_open
    checkbox_select.close()
    assert not checkbox_select.is_open


def test_checkbox_select_item_checkbox_select(checkbox_select):
    checkbox_select.fill({"Restarted": True, "Cancelled": True})
    assert checkbox_select.read() == {
        "Active This is a description": False,
        "Cancelled": True,
        "Paused": False,
        "Warning": False,
        "Restarted": True,
    }

    checkbox_select.fill(
        {"Cancelled": False, "Paused": False, "Warning": False, "Restarted": False}
    )

    assert checkbox_select.read() == {
        "Active This is a description": False,
        "Cancelled": False,
        "Paused": False,
        "Warning": False,
        "Restarted": False,
    }

    assert not checkbox_select.is_open
    with pytest.raises(SelectItemNotFound):
        checkbox_select.fill({"Non existing item": True})
    assert not checkbox_select.is_open



import re

from widgetastic.widget import ClickableMixin
from widgetastic.widget import ParametrizedLocator
from widgetastic.widget import ParametrizedView
from widgetastic.widget import View
from widgetastic.widget import Widget
from widgetastic.xpath import quote


class DonutLegendItem(ParametrizedView, ClickableMixin):
    PARAMETERS = ("label_text",)
    ROOT = ParametrizedLocator(
        ".//*[name()='text']" "/*[name()='tspan' and contains(., '{label_text}')]"
    )
    ALL_ITEMS = ".//*[name()='text']/*[name()='tspan']"
    LEGEND_ITEM_REGEX = re.compile(r"(.*?): ([\d]+)")

    @classmethod
    def _get_legend_item(cls, text):
        match = cls.LEGEND_ITEM_REGEX.match(text)
        if match:
            return match.group(1), match.group(2)
        else:
            return text, None

    @property
    def label(self):
        """Returns the label of a DonutLegendItem as a string"""
        return self._get_legend_item(self.browser.text(self))[0]

    @property
    def value(self):
        """Returns the value of a DonutLegendItem as a string"""
        return self._get_legend_item(self.browser.text(self))[1]

    @classmethod
    def all(cls, browser):
        """Returns a list of all items"""
        return [(browser.text(el),) for el in browser.elements(cls.ALL_ITEMS)]


class DonutLegend(View):
    ROOT = ".//*[name()='g'][2]"

    item = ParametrizedView.nested(DonutLegendItem)

    @property
    def all_items(self):
        """Returns a list of all items, arranged as {label: value}"""
        result = []
        for i in self.item:
            result.append({"label": i.label, "value": i.value})
        return result


class DonutCircle(View):
    ROOT = ".//div[*[name()='svg'][*[name()='text'] and not(*[name()='rect'])]]"
    LABELS_LOCATOR = "./*[name()='svg']/*[name()='text']/*[name()='tspan']"

    @property
    def labels(self):
        """Returns a list of labels"""
        return [self.browser.text(elem) for elem in self.browser.elements(self.LABELS_LOCATOR)]


class DonutChart(View):
    """
    Represents the Donut Chart
    from Patternfly 4 (https://www.patternfly.org/v4/documentation/react/charts/chartdonut)
    """

    ROOT = ParametrizedLocator("{@locator}")
    BASE_LOCATOR = ".//div[@id={}]"

    donut = View.nested(DonutCircle)
    legend = View.nested(DonutLegend)

    def __init__(self, parent, id=None, locator=None, logger=None):
        """Create the widget"""
        Widget.__init__(self, parent, logger=logger)
        if id:
            self.locator = self.BASE_LOCATOR.format(quote(id))
        elif locator:
            self.locator = locator
        else:
            raise TypeError("You need to specify either id or locator")



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Menu
from widgetastic_patternfly4 import MenuItemNotFound

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/menu"


@pytest.fixture
def menu(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-menu-option-single-select']"
        menu = Menu(locator=".//div[contains(@class, 'pf-c-menu')]")

    return TestView(browser).menu


@pytest.fixture
def multi_select_menu(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-menu-option-multi-select']"
        menu = Menu(locator=".//div[contains(@class, 'pf-c-menu')]")

    return TestView(browser).menu


def test_menu_is_displayed(menu):
    assert menu.is_displayed


def test_menu_items(menu):
    assert set(menu.items) == {"Option 1", "Option 2", "Option 3"}
    assert menu.has_item("Option 1")
    assert not menu.has_item("Non existing item")
    assert menu.item_enabled("Option 2")


def test_menu_item_select(menu):
    menu.fill("Option 1")
    assert menu.selected_items == ["Option 1"]
    assert menu.is_open
    with pytest.raises(MenuItemNotFound):
        menu.fill("Non existing item")
    menu.fill("Option 3")
    assert menu.selected_items == ["Option 3"]


def test_menu_multi_item_select(multi_select_menu):
    menu = multi_select_menu

    assert not menu.selected_items
    menu.fill(["Option 1", "Option 2", "Option 3"])
    assert menu.selected_items == ["Option 1", "Option 2", "Option 3"]



from .alert import Alert
from .breadcrumb import BreadCrumb
from .bulletchart import BulletChart
from .button import Button
from .card import Card
from .card import CardCheckBox
from .card import CardForCardGroup
from .card import CardGroup
from .chipgroup import CategoryChipGroup
from .chipgroup import Chip
from .chipgroup import ChipGroup
from .chipgroup import ChipGroupToolbar
from .chipgroup import ChipGroupToolbarCategory
from .chipgroup import ChipReadOnlyError
from .chipgroup import StandAloneChipGroup
from .contextselector import ContextSelector
from .donutchart import DonutChart
from .dropdown import Dropdown
from .dropdown import DropdownDisabled
from .dropdown import DropdownItemDisabled
from .dropdown import DropdownItemNotFound
from .dropdown import GroupDropdown
from .dropdown import SplitButtonDropdown
from .formselect import FormSelect
from .formselect import FormSelectDisabled
from .formselect import FormSelectOptionDisabled
from .formselect import FormSelectOptionNotFound
from .menu import CheckboxMenu
from .menu import Menu
from .menu import MenuItemDisabled
from .menu import MenuItemNotFound
from .modal import Modal
from .navigation import Navigation
from .navigation import NavSelectionNotFound
from .optionsmenu import OptionsMenu
from .pagination import CompactPagination
from .pagination import Pagination
from .pagination import PaginationNavDisabled
from .piechart import PieChart
from .popover import Popover
from .progress import Progress
from .select import CheckboxSelect
from .select import Select
from .select import SelectItemDisabled
from .select import SelectItemNotFound
from .switch import Switch
from .switch import SwitchDisabled
from .table import ColumnNotExpandable
from .table import CompoundExpandableTable
from .table import ExpandableTable
from .table import PatternflyTable
from .table import RowNotExpandable
from .tabs import Tab

__all__ = [
    "Alert",
    "BreadCrumb",
    "Button",
    "BulletChart",
    "CheckboxMenu",
    "CheckboxSelect",
    "Chip",
    "ChipReadOnlyError",
    "ChipGroup",
    "ChipGroupToolbar",
    "ChipGroupToolbarCategory",
    "StandAloneChipGroup",
    "Card",
    "CardGroup",
    "CardForCardGroup",
    "CardCheckBox",
    "ColumnNotExpandable",
    "CompactPagination",
    "CompoundExpandableTable",
    "CategoryChipGroup",
    "DonutChart",
    "Dropdown",
    "DropdownDisabled",
    "DropdownItemDisabled",
    "DropdownItemNotFound",
    "GroupDropdown",
    "SplitButtonDropdown",
    "FormSelect",
    "FormSelectDisabled",
    "FormSelectOptionDisabled",
    "FormSelectOptionNotFound",
    "Menu",
    "MenuItemDisabled",
    "MenuItemNotFound",
    "Modal",
    "Navigation",
    "NavSelectionNotFound",
    "OptionsMenu",
    "Pagination",
    "PaginationNavDisabled",
    "PieChart",
    "Popover",
    "Progress",
    "Select",
    "SelectItemDisabled",
    "SelectItemNotFound",
    "ContextSelector",
    "Switch",
    "SwitchDisabled",
    "ExpandableTable",
    "PatternflyTable",
    "RowNotExpandable",
    "Tab",
]



import re

import pytest
from widgetastic.exceptions import WidgetOperationFailed
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import BreadCrumb

TESTING_PAGE_URL = (
    "https://patternfly-docs-ouia.netlify.app/documentation/react/components/breadcrumb"  # noqa
)


def test_breadcrumb(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-breadcrumb-ouia']"
        breadcrumb = BreadCrumb("basic")

    view = TestView(browser)
    assert view.breadcrumb.is_displayed
    assert len(view.breadcrumb.locations) == 4
    assert view.breadcrumb.locations[0].lower() == "section home"
    assert view.breadcrumb.read().lower() == "section landing"
    view.breadcrumb.click_location(view.breadcrumb.locations[0])
    view.breadcrumb.click_location("title", partial=True)

    failing_location = "definitely not in the example page"

    # exception + message on full match
    exception_match = re.escape(
        f'Breadcrumb location "{failing_location}" '
        f"not found within locations: {view.breadcrumb.locations}"
    )
    with pytest.raises(WidgetOperationFailed, match=exception_match):
        view.breadcrumb.click_location(failing_location)

    # exception + message on partial match
    exception_match = re.escape(
        f'Breadcrumb location "{failing_location}" '
        "not found with partial match "
        f"within locations: {view.breadcrumb.locations}"
    )
    with pytest.raises(WidgetOperationFailed, match=exception_match):
        view.breadcrumb.click_location(failing_location, partial=True)



import re

from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Text
from widgetastic.widget import View
from widgetastic.xpath import quote


class Legend:
    """Represents Legend of chart."""

    def __init__(self, label, value=None, color=None, element=None):
        self.label = label
        self.value = value
        self.color = color
        self.element = element

    def click(self):
        self.element.click()

    def __repr__(self):
        return f"Legend({self.label}: {self.value})"

    def __gt__(self, leg):
        return self.__class__ == leg.__class__ and self.value > leg.value

    def __eq__(self, leg):
        return (
            self.__class__ == self.__class__ and self.label == leg.label and self.value == leg.value
        )


class DataPoint(Legend):
    """Represents DataPoint on chart."""

    def __repr__(self):
        return f"DataPoint({self.label}: {self.value})"


class BulletChart(View):
    """Represents the Patternfly Bullet Chart.

    https://www.patternfly.org/v4/documentation/react/charts/chartbullet

    Args:
        id: If you want to look the input up by id, use this parameter, pass the id.
        locator: If you have specific locator else it will take pf-chart.
        offset_denominator: Denominator for offset value calculation.
    """

    ROOT = ParametrizedLocator("{@locator}")
    DEFAULT_LOCATOR = ".//div[contains(@class, 'chartbullet')]"
    LEGEND_ICON = (
        ".//*[name()='g' or name()='svg']/*[name()='rect']/following-sibling::*[name()='path']"
    )
    LEGEND_TEXT = ".//*[name()='text' and contains(@id, 'legend-labels')]"
    LEGEND_ITEM_REGEX = re.compile(r"(\d+)\s(\w.*)|(\w.*)\s(\d+)")

    ITEMS = ".//*[name()='g']/*[name()='path' and not(contains(@style, 'type:square'))]"
    TOOLTIP_REGEX = re.compile(r"(.*?): ([\d]+)")
    APPLY_OFFSET = True

    tooltip = Text(
        ".//*[name()='svg' and contains(@aria-labelledby, 'victory-container')]/"
        "following-sibling::div[contains(@style, 'z-index')]/*[name()='svg']"
    )

    def __init__(self, parent=None, id=None, locator=None, logger=None, *args, **kwargs):
        View.__init__(self, parent=parent, logger=logger)
        if id:
            self.locator = ".//div[@id={}]".format(quote(id))
        elif locator:
            self.locator = locator
        else:
            self.locator = self.DEFAULT_LOCATOR

        self.args = args
        self.kwargs = kwargs

    def _offsets(self, el):
        """Calculate offset. Need to set offset with try and error method."""
        offset_denominator = self.kwargs.pop("offset_denominator", 2.5)
        size = self.browser.size_of(el)
        width = size.width
        height = size.height
        dx = int(width / offset_denominator) if (width > 10 and height > 10) else 0
        dy = int(height / offset_denominator) if (width > 10 and height > 10) else 0
        return dx, dy

    @property
    def legends(self):
        """Return all Legend objects."""

        br = self.browser
        _data = []
        for (icon, label_el) in zip(br.elements(self.LEGEND_ICON), br.elements(self.LEGEND_TEXT)):
            label_text = br.text(label_el)
            match = self.LEGEND_ITEM_REGEX.match(label_text)
            if match:
                left_value, left_label, right_label, right_value = match.groups()
            label = right_label or left_label if match else label_text
            value = int(right_value or left_value) if match else None
            _data.append(
                Legend(
                    label=label,
                    value=value,
                    color=icon.value_of_css_property("fill"),
                    element=label_el,
                )
            )
        return _data

    @property
    def legend_names(self):
        """Return all legend names."""
        return [leg.label for leg in self.legends]

    def get_legend(self, label):
        """Get specific Legend object.

        Args:
            label: Name of legend label.
        """
        try:
            return next(leg for leg in self.legends if leg.label == label)
        except StopIteration:
            return None

    @property
    def data(self):
        """Read graph and returns all Data Point objects."""
        _data = []
        # focus away from graph
        self.parent_browser.move_to_element("//body")

        for el in self.browser.elements(self.ITEMS):
            self.browser.move_to_element(el)

            if self.APPLY_OFFSET:
                dx, dy = self._offsets(el)
                self.browser.move_by_offset(dx, dy)

            match = self.TOOLTIP_REGEX.match(self.tooltip.text)
            if match:
                _data.append(
                    DataPoint(
                        label=match.groups()[0],
                        value=int(match.groups()[1]),
                        color=el.value_of_css_property("fill"),
                        element=el,
                    )
                )
        return _data

    def get_data_point(self, label):
        """Get specific data point object.

        Args:
            label: Name of respective data point label.
        """
        try:
            return next(dp for dp in self.data if dp.label == label)
        except StopIteration:
            return None

    def read(self):
        """Read graph and returns label, value dict."""
        return {dp.label: dp.value for dp in self.data}



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import Alert

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/alert"

ALERT_TYPES = ["success", "danger", "warning", "info"]


@pytest.fixture(params=ALERT_TYPES)
def alert(browser, request):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-alert-ouia']"
        alert = Alert(request.param)

    view = TestView(browser)
    return view.alert


def test_alert_is_displayed(alert):
    assert alert.is_displayed


def test_alert_title(alert):
    alert_type = alert.type if alert.type != "error" else "danger"
    assert alert.title == f"{alert_type.capitalize()} alert title"



import os
import subprocess
from urllib.request import urlopen

import pytest
from selenium import webdriver
from wait_for import wait_for
from widgetastic.browser import Browser


OPTIONS = {"firefox": webdriver.FirefoxOptions(), "chrome": webdriver.ChromeOptions()}


def pytest_addoption(parser):
    parser.addoption(
        "--browser-name",
        help="name of the browser",
        choices=("firefox", "chrome"),
        default="firefox",
    )

    parser.addoption("--force-host", default=None, help="force a selenium hostname")


@pytest.fixture(scope="session")
def browser_name(pytestconfig):
    return os.environ.get("BROWSER") or pytestconfig.getoption("--browser-name")


@pytest.fixture(scope="session")
def selenium_url(pytestconfig, worker_id):
    forced_host = pytestconfig.getoption("--force-host")

    if forced_host is None:
        oktet = 1 if worker_id == "master" else int(worker_id.lstrip("gw")) + 1
        host = f"127.0.0.{oktet}"
        ps = subprocess.run(
            [
                "podman",
                "run",
                "--rm",
                "-d",
                "-p",
                f"{host}:4444:4444",
                "-p",
                f"{host}:5999:5999",
                "--shm-size=2g",
                "quay.io/redhatqe/selenium-standalone:ff_91.8.0esr_chrome_101.0.4951.41",
            ],
            stdout=subprocess.PIPE,
        )

        yield f"http://{host}:4444"
        container_id = ps.stdout.decode("utf-8").strip()
        subprocess.run(["podman", "kill", container_id], stdout=subprocess.DEVNULL)
    else:
        yield f"http://{forced_host}:4444"


@pytest.fixture(scope="session")
def wait_for_selenium(selenium_url):
    wait_for(lambda: urlopen(selenium_url), timeout=180, handle_exception=True)


@pytest.fixture(scope="session")
def selenium(browser_name, wait_for_selenium, selenium_url):
    driver = webdriver.Remote(command_executor=selenium_url, options=OPTIONS[browser_name.lower()])
    driver.maximize_window()
    yield driver
    driver.quit()


@pytest.fixture(scope="module")
def browser(selenium, request):
    selenium.get(request.module.TESTING_PAGE_URL)
    yield Browser(selenium)
    selenium.refresh()



from widgetastic.exceptions import NoSuchElementException

from .dropdown import Dropdown
from .dropdown import DropdownItemDisabled
from .dropdown import DropdownItemNotFound


class SelectItemDisabled(DropdownItemDisabled):
    pass


class SelectItemNotFound(DropdownItemNotFound):
    pass


class BaseSelect:
    """Represents the Patternfly Select.

    https://www.patternfly.org/v4/documentation/react/components/select
    """

    BUTTON_LOCATOR = "./button"
    ITEMS_LOCATOR = (
        ".//ul[@class='pf-c-select__menu']/li[contains(@class, 'pf-c-select__menu-wrapper')]"
    )
    ITEM_LOCATOR = ".//*[contains(@class, 'pf-c-select__menu-item') and normalize-space(.)={}]"
    SELECTED_ITEM_LOCATOR = (
        ".//span[contains(@class, 'ins-c-conditional-filter') and normalize-space(.)={}]"
    )
    TEXT_LOCATOR = (
        ".//div[contains(@class, 'pf-c-select') and child::button[normalize-space(.)={}]]"
    )

    def item_element(self, item, close=True):
        """Returns a WebElement for given item name."""
        try:
            return super().item_element(item, close)
        except DropdownItemNotFound:
            raise SelectItemNotFound(
                "Item {!r} not found in {}. Available items: {}".format(
                    item, repr(self), self.items
                )
            )

    def item_select(self, item):
        """Opens the Select and selects the desired item.

        Args:
            item: Item to be selected

        Raises:
            SelectItemDisabled: if item is disabled
        """
        try:
            return super().item_select(item)
        except DropdownItemDisabled:
            raise SelectItemDisabled(
                'Item "{}" of {} is disabled\n'
                "The following items are available and enabled: {}".format(
                    item, repr(self), self.enabled_items
                )
            )

    def fill(self, value):
        """Fills a Select with a value."""
        self.item_select(value)

    def read(self):
        """Returns a string of the text of the selected option."""
        return self.browser.text(self.BUTTON_LOCATOR)


class Select(BaseSelect, Dropdown):
    DEFAULT_LOCATOR = './/div[contains(@class, "pf-c-select")][1]'


class BaseCheckboxSelect(BaseSelect):
    """Represents the Patternfly Checkbox Select.

    https://www.patternfly.org/v4/documentation/react/components/select
    """

    ITEMS_LOCATOR = ".//label[contains(@class, 'pf-c-select__menu-item')]"
    ITEM_LOCATOR = f"{ITEMS_LOCATOR}/span[starts-with(normalize-space(.), {{}})]/preceding-sibling::input"  # noqa

    def item_select(self, items, close=True):
        """Opens the Checkbox and selects the desired item.

        Args:
            item: Item to be selected
            close: Close the dropdown when finished
        """
        if not isinstance(items, (list, tuple, set)):
            items = [items]

        try:
            for item in items:
                element = self.item_element(item, close=False)
                if not self.browser.is_selected(element):
                    element.click()
        finally:
            if close:
                self.close()

    def item_deselect(self, items, close=True):
        """Opens the Checkbox and deselects the desired item.

        Args:
            item: Item to be selected
            close: Close the dropdown when finished
        """
        if not isinstance(items, (list, tuple, set)):
            items = [items]

        try:
            for item in items:
                element = self.item_element(item, close=False)
                if self.browser.is_selected(element):
                    element.click()
        finally:
            if close:
                self.close()

    def fill(self, items):
        """Fills a Checkbox with all items.
        Example dictionary: {"foo": True, "bar": False, "baz": True}

        Args:
            items: A dictionary containing what items to select (True) or deselect (False)
        """
        try:
            for item, value in items.items():
                if value:
                    self.item_select(item, close=False)
                else:
                    self.item_deselect(item, close=False)
        finally:
            self.close()

    def read(self):
        """Returns a dictionary containing the selected status as bools."""
        selected = {}
        with self.opened():
            for el in self.browser.elements(self.ITEMS_LOCATOR):
                item = self.browser.text(el)
                try:
                    # get the child element of the label
                    selected[item] = self.browser.element(
                        parent=el, locator="./input"
                    ).is_selected()
                except NoSuchElementException:
                    selected[item] = False

        return selected

    def _get_items(self, close=False):
        """Returns a list of all checkbox items as strings.

        Args:
            close: Close the dropdown when finished
        """
        self.open()
        result = [self.browser.text(el) for el in self.browser.elements(self.ITEMS_LOCATOR)]

        if close:
            self.close()

        return result

    @property
    def items(self):
        """Returns a list of all CheckboxSelect items as strings."""
        return self._get_items(close=True)


class CheckboxSelect(BaseCheckboxSelect, Dropdown):
    DEFAULT_LOCATOR = './/div[contains(@class, "pf-c-select")][1]'



import pytest
from widgetastic.widget import Text

from widgetastic_patternfly4.button import Button
from widgetastic_patternfly4.modal import Modal
from widgetastic_patternfly4.modal import ModalItemNotFound

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/modal"


@pytest.fixture()
def modal(browser):
    show_modal = Button(browser, "Show modal")
    show_modal.click()
    modal = Modal(browser)
    yield modal
    if modal.is_displayed:
        modal.close()


class CustomModal(Modal):
    """Model use as view and enhance with widgets"""

    custom_body = Text(".//div[contains(@class, 'pf-c-modal-box__body')]")


def test_title(modal):
    assert modal.title


def test_body(modal):
    body = modal.body
    assert body.text.startswith("Lorem")


def test_close(modal):
    modal.close()
    assert not modal.is_displayed


def test_footer_items(modal):
    items = modal.footer_items
    assert len(items) == 2
    assert "Cancel" in items
    assert "Confirm" in items


def test_footer_item(modal):
    item = modal.footer_item("Confirm")
    assert item.text == "Confirm"
    item.click()
    assert not modal.is_displayed


def test_footer_item_invalid(modal):
    try:
        modal.footer_item("INVALID")
    except ModalItemNotFound:
        assert True
    else:
        pytest.fail("ModalItemNotFound exception expected.")


def test_modal_as_view(browser, modal):
    view = CustomModal(browser)
    assert view.is_displayed
    assert view.custom_body.text == modal.body.text



from wait_for import wait_for_decorator
from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import View


class Tab(View):
    """Represents the Patternfly Tab widget.

    Selects itself automatically when any child widget gets accessed, ensuring that the widget is
    visible.

    https://www.patternfly.org/v4/documentation/react/components/tabs
    """

    # The text on the tab. Can be omitted if it is the same as the tab class name capitalized
    TAB_NAME = None

    # Locator of the Tab selector
    TAB_LOCATOR = ParametrizedLocator(
        './/div[contains(@class, "pf-c-tabs")]/ul'
        "/li[button[normalize-space(.)={@tab_name|quote}]]"
    )

    ROOT = ParametrizedLocator(
        ".//section[@aria-labelledby=string("
        "preceding-sibling::div/ul/li/button[normalize-space(.)={@tab_name|quote}]/@id)]"
        "|"
        ".//section[@id=string(../preceding-sibling::div/ul/li"
        "/button[normalize-space(.)={@tab_name|quote}]/@aria-controls)]"
    )

    @property
    def tab_name(self):
        """Returns the tab name as a string."""
        return self.TAB_NAME or type(self).__name__.replace("_", " ").capitalize()

    def is_active(self):
        """Returns a boolean detailing of the tab is active."""
        return "pf-m-current" in self.parent_browser.classes(self.TAB_LOCATOR)

    @property
    def is_displayed(self):
        """Returns a boolean detailing of the tab is displayed."""
        return self.parent_browser.is_displayed(self.TAB_LOCATOR)

    def click(self):
        """Clicks the tab."""
        return self.parent_browser.click(self.TAB_LOCATOR)

    def select(self):
        """Selects the tab (checks if active already first)."""
        if not self.is_active():
            self.logger.info("Opening the tab %s", self.tab_name)

            @wait_for_decorator(timeout=3)
            def _click():
                self.click()
                return self.is_active()

    def child_widget_accessed(self, widget):
        # Select the tab
        self.select()

    def __repr__(self):
        return "<Tab {!r}>".format(self.tab_name)



from contextlib import contextmanager

from wait_for import wait_for_decorator
from widgetastic.exceptions import NoSuchElementException
from widgetastic.exceptions import UnexpectedAlertPresentException
from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Checkbox
from widgetastic.widget import Widget
from widgetastic.xpath import quote


class DropdownDisabled(Exception):
    pass


class DropdownItemDisabled(Exception):
    pass


class DropdownItemNotFound(Exception):
    pass


class BaseDropdown:
    """Represents the Patternfly dropdown.

    https://www.patternfly.org/v4/documentation/react/components/dropdown

    Args:
        text: Text of the button, can be the inner text or the title attribute.

    """

    BUTTON_LOCATOR = ".//button[contains(@class, 'pf-c-dropdown__toggle')]"
    ITEMS_LOCATOR = ".//ul[contains(@class, 'pf-c-dropdown__menu')]/li"
    ITEM_LOCATOR = ".//*[contains(@class, 'pf-c-dropdown__menu-item') and normalize-space(.)={}]"

    @contextmanager
    def opened(self):
        """A context manager to open and then close a Dropdown."""
        self.open()
        yield
        self.close()

    @property
    def is_enabled(self):
        """Returns if the dropdown itself is enabled and therefore interactive."""
        return "disabled" not in self.browser.classes(self.BUTTON_LOCATOR)

    def _verify_enabled(self):
        if not self.is_enabled:
            raise DropdownDisabled(
                '{} "{}" is not enabled'.format(
                    type(self).__name__, getattr(self, "text", self.locator)
                )
            )

    @property
    def is_open(self):
        """Returns True if the Dropdown is open"""
        return "pf-m-expanded" in self.browser.classes(self)

    def open(self):
        """Opens a dropdown."""
        self._verify_enabled()
        if self.is_open:
            return

        @wait_for_decorator(timeout=3)
        def _click():
            self.browser.click(self.BUTTON_LOCATOR)
            return self.is_open

    def close(self, ignore_nonpresent=False):
        """Close the dropdown

        Args:
            ignore_nonpresent: Will ignore exceptions due to disabled or missing dropdown
        """
        try:
            self._verify_enabled()
            if self.is_open:
                self.browser.click(self.BUTTON_LOCATOR)
        except (NoSuchElementException, DropdownDisabled):
            if ignore_nonpresent:
                self.logger.info("%r hid so it was not possible to close it. But ignoring.", self)
            else:
                raise

    @property
    def items(self):
        """Returns a list of all dropdown items as strings."""
        with self.opened():
            result = [self.browser.text(el) for el in self.browser.elements(self.ITEMS_LOCATOR)]
        return result

    def has_item(self, item):
        """Returns whether the items exists.

        Args:
            item: item name

        Returns:
            Boolean - True if enabled, False if not.
        """
        return item in self.items

    def item_element(self, item, close=True, **kwargs):
        """Returns a WebElement for given item name."""
        try:
            self.open()
            result = self.browser.element(self.ITEM_LOCATOR.format(quote(item)), **kwargs)
            if close:
                self.close()
            return result
        except NoSuchElementException:
            try:
                items = self.items
            except NoSuchElementException:
                items = []
            if items:
                items_string = "These items are present: {}".format("; ".join(items))
            else:
                items_string = "The dropdown is probably not present"
            raise DropdownItemNotFound("Item {!r} not found. {}".format(item, items_string))

    def item_enabled(self, item, close=True, **kwargs):
        """Returns whether the given item is enabled.

        Args:
            item: Name of the item.

        Returns:
            Boolean - True if enabled, False if not.
        """
        self._verify_enabled()
        el = self.item_element(item, close=False, **kwargs)
        is_el_enabled = "pf-m-disabled" not in self.browser.classes(el)
        if close:
            self.close()
        return is_el_enabled

    def item_select(self, item, handle_alert=None, **kwargs):
        """Opens the dropdown and selects the desired item.

        Args:
            item: Item to be selected
            handle_alert: How to handle alerts. None - no handling, True - confirm, False - dismiss.

        Raises:
            DropdownItemDisabled
        """
        self.logger.info("Selecting %r", item)
        try:
            if not self.item_enabled(item, close=False, **kwargs):
                raise DropdownItemDisabled(
                    'Item "{}" of {} "{}" is disabled\n'
                    "The following items are available: {}".format(
                        item,
                        type(self).__name__.lower(),
                        getattr(self, "text", None) or self.locator,
                        ";".join(self.items),
                    )
                )
            self.browser.click(
                self.item_element(item, close=False, **kwargs), ignore_ajax=handle_alert is not None
            )
            if handle_alert is not None:
                self.browser.handle_alert(cancel=not handle_alert, wait=10.0)
                self.browser.plugin.ensure_page_safe()
        finally:
            try:
                self.close(ignore_nonpresent=True)
            except UnexpectedAlertPresentException:
                self.logger.warning("There is an unexpected alert present.")
                pass

    @property
    def button_text(self):
        """Returns a string of the current dropdown name."""
        return self.browser.text(self.BUTTON_LOCATOR)

    def read(self):
        return self.button_text

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, getattr(self, "text", None) or self.locator)


class Dropdown(BaseDropdown, Widget):
    ROOT = ParametrizedLocator("{@locator}")
    TEXT_LOCATOR = (
        ".//div[contains(@class, 'pf-c-dropdown') and child::button[normalize-space(.)={}]]"
    )
    DEFAULT_LOCATOR = './/div[contains(@class, "pf-c-dropdown")][1]'

    def __init__(self, parent, text=None, locator=None, logger=None):
        super().__init__(parent, logger=logger)
        if locator and text:
            raise ValueError("Either text or locator should be provided")
        if text:
            self.locator = self.TEXT_LOCATOR.format(quote(text))
        elif locator:
            self.locator = locator
        else:
            self.locator = self.DEFAULT_LOCATOR


class BaseGroupDropdown:
    """Dropdown with grouped items in it."""

    ITEMS_LOCATOR = ".//section[@class='pf-c-dropdown__group']/ul/li"
    GROUPS_LOCATOR = ".//section[@class='pf-c-dropdown__group']/h1"
    GROUP_LOCATOR = ".//section[@class='pf-c-dropdown__group'][h1[normalize-space(.)={}]]"

    @property
    def groups(self):
        """Returns a list of all group names as strings."""
        with self.opened():
            result = [self.browser.text(el) for el in self.browser.elements(self.GROUPS_LOCATOR)]
        return result

    def item_element(self, item, group_name=None, close=True):
        """Returns a WebElement for given item name."""
        self.open()
        try:
            kwargs = (
                {"parent": self.browser.element(self.GROUP_LOCATOR.format(quote(group_name)))}
                if group_name
                else {}
            )
        except NoSuchElementException:
            raise DropdownItemNotFound(
                'Following group "{}" not found. Available are: {}'.format(group_name, self.groups)
            )
        return super().item_element(item, close=close, **kwargs)

    def item_select(self, item, group_name=None, handle_alert=None):
        """Opens the dropdown and selects the desired item. Implemented only for proper kwargs
        suggestions.

        Args:
            item: Item to be selected
            group_name: name of a group to search in. If not provided - all groups will be checked.
            handle_alert: How to handle alerts. None - no handling, True - confirm, False - dismiss.

        Raises:
            DropdownItemDisabled
        """
        return super().item_select(item, handle_alert=handle_alert, group_name=group_name)


class GroupDropdown(BaseGroupDropdown, Dropdown):
    pass


class BaseSplitButtonDropdown:
    """Represents the Patternfly Split Button Dropdown.

    https://www.patternfly.org/v4/documentation/react/components/dropdown#split-button-with-text
    """

    toggle_check = Checkbox(locator=".//input[@type='checkbox']")

    def check(self):
        """Check toggle checkbox."""
        return self.toggle_check.fill(True)

    def uncheck(self):
        """Uncheck toggle checkbox."""
        return self.toggle_check.fill(False)

    @property
    def selected(self):
        """Returns selected or not"""
        return self.toggle_check.selected

    def read(self):
        return self.browser.text(self)


class SplitButtonDropdown(BaseSplitButtonDropdown, Dropdown):
    pass



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Dropdown
from widgetastic_patternfly4 import DropdownItemDisabled
from widgetastic_patternfly4 import DropdownItemNotFound
from widgetastic_patternfly4 import GroupDropdown
from widgetastic_patternfly4 import SplitButtonDropdown

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/dropdown"


@pytest.fixture
def view(browser):
    class TestView(View):
        ROOT = "(.//div[@id='ws-react-c-dropdown-basic'])[1]"
        dropdown_txt_locator = Dropdown("Dropdown")
        dropdown_custom_locator = Dropdown(locator=".//div[contains(@class, 'pf-c-dropdown')]")
        dropdown_default_locator = Dropdown()

    return TestView(browser)


@pytest.fixture(
    params=["dropdown_txt_locator", "dropdown_custom_locator", "dropdown_default_locator"]
)
def dropdown(view, request):
    return getattr(view, request.param)


@pytest.fixture()
def group_dropdown(browser):
    return GroupDropdown(
        browser,
        locator=(
            ".//div[@id='ws-react-c-dropdown-with-groups']"
            "/div[contains(@class, 'pf-c-dropdown')]"
        ),
    )


@pytest.fixture(
    params=["ws-react-c-dropdown-split-button", "ws-react-c-dropdown-split-button-with-text"],
    ids=["without_text", "with_text"],
)
def split_button_dropdown(request, browser):
    return (
        SplitButtonDropdown(
            browser, locator=f".//div[@id='{request.param}']/div[contains(@class, 'pf-c-dropdown')]"
        ),
        request.param,
    )


def test_dropdown_is_displayed(dropdown):
    assert dropdown.is_displayed


def test_enabled_dropdown(dropdown):
    assert dropdown.is_enabled


def test_dropdown_items(dropdown):
    assert dropdown.items == [
        "Link",
        "Action",
        "Disabled link",
        "Disabled action",
        "",
        "Separated link",
        "Separated action",
    ]
    assert dropdown.has_item("Action")
    assert not dropdown.has_item("Non existing items")
    assert dropdown.item_enabled("Action")
    assert not dropdown.item_enabled("Disabled link")


def test_dropdown_open(dropdown):
    assert not dropdown.is_open
    dropdown.open()
    assert dropdown.is_open
    dropdown.close()
    assert not dropdown.is_open


def test_dropdown_item_select(dropdown):
    dropdown.item_select("Action")
    assert not dropdown.is_open
    with pytest.raises(DropdownItemDisabled):
        dropdown.item_select("Disabled link")
    with pytest.raises(DropdownItemNotFound):
        dropdown.item_select("Non existing items")


def test_group_dropdown(group_dropdown):
    assert group_dropdown.is_displayed
    assert group_dropdown.is_enabled
    assert group_dropdown.items == [
        "Link",
        "Action",
        "Group 2 link",
        "Group 2 action",
        "Group 3 link",
        "Group 3 action",
    ]
    assert group_dropdown.has_item("Group 2 link")
    assert group_dropdown.item_enabled("Group 3 action")
    assert group_dropdown.groups == ["Group 2", "Group 3"]
    group_dropdown.item_select("Link")
    group_dropdown.item_select("Group 3 link", group_name="Group 3")
    with pytest.raises(DropdownItemNotFound):
        group_dropdown.item_select("Group 3 link", group_name="Group 2")


def test_split_button_dropdown(split_button_dropdown):
    dropdown, dropdown_type = split_button_dropdown
    assert dropdown.is_displayed
    assert dropdown.is_enabled
    assert dropdown.has_item("Action")
    dropdown.item_select("Action")
    assert dropdown.check()
    assert dropdown.selected
    expected_text = "10 selected" if "with-text" in dropdown_type else ""
    assert dropdown.read() == expected_text

    assert dropdown.uncheck()
    assert not dropdown.selected



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import DropdownItemNotFound
from widgetastic_patternfly4.ouia import OptionsMenu

TESTING_PAGE_URL = (
    "https://patternfly-docs-ouia.netlify.app/documentation/react/components/optionsmenu"  # noqa
)


@pytest.fixture
def options_menu(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-optionsmenu-ouia']"
        options_menu = OptionsMenu("Simple Options Menu")

    return TestView(browser).options_menu


def test_options_menu_is_displayed(options_menu):
    assert options_menu.is_displayed


def test_enabled_options_menu(options_menu):
    assert options_menu.is_enabled


def test_options_menu_items(options_menu):
    assert options_menu.items == ["Option 1", "Option 2", "Option 3"]
    assert options_menu.has_item("Option 2")
    assert not options_menu.has_item("Non existing items")
    assert options_menu.item_enabled("Option 1")


def test_options_menu_open(options_menu):
    assert not options_menu.is_open
    options_menu.open()
    assert options_menu.is_open
    options_menu.close()
    assert not options_menu.is_open


def test_options_menu_item_select(options_menu):
    options_menu.item_select("Option 2")
    assert options_menu.selected_items[0] == "Option 2"
    assert not options_menu.is_open
    with pytest.raises(DropdownItemNotFound):
        options_menu.item_select("Non existing items")



import re

from widgetastic_patternfly4 import BulletChart


class PieChart(BulletChart):
    """
    Represents the Pie Chart
    from Patternfly 4 (https://www.patternfly.org/v4/documentation/react/charts/chartpie)
    """

    APPLY_OFFSET = False
    LEGEND_ITEM_REGEX = re.compile(r"(\d+): (\w.*)|(\w.*): (\d+)")



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import FormSelectOptionNotFound
from widgetastic_patternfly4.ouia import FormSelect

TESTING_PAGE_URL = (
    "https://patternfly-docs-ouia.netlify.app/documentation/react/components/formselect"  # noqa
)


@pytest.fixture
def view(browser):
    class FormSelectTestView(View):
        ROOT = ".//div[@id='ws-react-c-formselect-ouia']"
        input = FormSelect("FormSelect Input")

    return FormSelectTestView(browser)


def test_formselect_visibility(view):
    assert view.input.is_displayed


def test_formselect_enablement(view):
    assert view.input.is_enabled


def test_formselect_validity(view):
    assert view.input.is_valid


def test_formselect_value(view):
    assert len(view.input.all_options) == 7
    assert len(view.input.all_enabled_options) == 6
    assert "Mrs" in view.input.all_options
    view.input.fill("Mrs")
    assert view.input.read() == "Mrs"


def test_formselect_option_enablement(view):
    expected_enabled_options = {"Mr", "Miss", "Mrs", "Ms", "Dr", "Other"}
    expected_disabled_options = {"Please Choose"}
    assert set(view.input.all_enabled_options) == expected_enabled_options
    assert expected_disabled_options not in set(view.input.all_enabled_options)


def test_formselect_fill_nonexistent_option(view):
    with pytest.raises(FormSelectOptionNotFound):
        view.input.fill("foo")



from selenium.webdriver.support.ui import Select
from widgetastic.exceptions import NoSuchElementException
from widgetastic.widget import GenericLocatorWidget


class FormSelectDisabled(Exception):
    """Raised if the FormSelect is Disabled."""

    pass


class FormSelectOptionDisabled(Exception):
    """Raise if the specific option chosen is disabled."""

    pass


class FormSelectOptionNotFound(Exception):
    """Raised if the specific form option is not found in the form select."""

    pass


class BaseFormSelect:
    """Represents the Patternfly FormSelect.

    https://www.patternfly.org/v4/documentation/react/components/formselect
    """

    ALL_OPTIONS_LOCATOR = ".//option"
    PARENT_OPTION_GROUP = "./parent::optgroup"

    @property
    def is_enabled(self):
        """Returns whether the FormSelect itself is enabled and therefore interactive."""
        return self.browser.get_attribute("disabled", self) is None

    @property
    def is_valid(self):
        """Returns whether the FormSelect has valid option selected - not highlighted with red
        color and special icon.
        """
        return self.browser.get_attribute("aria-invalid", self) == "false"

    @property
    def all_options(self):
        """Returns a list of all the options in the FormSelect."""
        return [self.browser.text(el) for el in self.browser.elements(self.ALL_OPTIONS_LOCATOR)]

    @property
    def all_enabled_options(self):
        """Returns a list of all enabled options in the FormSelect.

        Options might be disabled by `disabled` attr in their WebElement or in parent optgroup
        element (if it exists)."""
        result = []
        all_option_elements = self.browser.elements(self.ALL_OPTIONS_LOCATOR)
        for el in all_option_elements:
            try:
                group_disabled = (
                    self.browser.get_attribute("disabled", self.PARENT_OPTION_GROUP, parent=el)
                    is not None
                )
            except NoSuchElementException:
                group_disabled = False
            element_disabled = self.browser.get_attribute("disabled", el) is not None
            if not group_disabled and not element_disabled:
                result.append(self.browser.text(el))
        return result

    @property
    def _select_element(self):
        return Select(self.__element__())

    def fill(self, value):
        """Select desired option in FormSelect.

        Raises:
            FormSelectDisabled: if FormSelect is disabled
            FormSelectOptionDisabled: if option or entire optgroup is disabled
            FormSelectOptionNotFound: if option not found
        """
        if not self.is_enabled:
            raise FormSelectDisabled("{} is not enabled".format(repr(self)))
        if value not in self.all_options:
            raise FormSelectOptionNotFound(
                'Option "{}" not found in {}. Available options: {}'.format(
                    value, repr(self), self.all_options
                )
            )
        elif value not in self.all_enabled_options:
            raise FormSelectOptionDisabled(
                'Option "{}" is disabled in {}. Enabled options are: {}'.format(
                    value, repr(self), self.all_enabled_options
                )
            )
        self._select_element.select_by_visible_text(value)

    def read(self):
        """Returns selected option."""
        return self.browser.text(self._select_element.first_selected_option)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.locator)


class FormSelect(BaseFormSelect, GenericLocatorWidget):
    pass



from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Widget


class BasePopover:
    """Represents a popover.
    https://www.patternfly.org/v4/components/popover/

    For the popover to work, the base view should include a property
    popover = Popover()
    As there is supposed to only be one popover active at a time, this property
    should work, interacting with the active popover.
    """

    TITLE = """.//*[self::h1 or self::h2 or self::h3 or self::h4 or self::h5 or self::h6
                and contains(@class, "pf-c-title")]"""
    BODY = './/div[contains(@class, "pf-c-popover__body")]'
    FOOTER = ".//footer[contains(@class, 'pf-c-popover__footer')]"
    CLOSE = ".//button[@aria-label='Close']"

    def read(self):
        """Returns the text of the body of the popover as a string."""
        return self.body

    @property
    def _raw_title_el(self):
        return self.browser.element(self.TITLE)

    @property
    def _raw_description_el(self):
        return self.browser.element(self.BODY)

    @property
    def _raw_footer_el(self):
        return self.browser.element(self.TITLE)

    @property
    def title(self):
        """Returns the title text of the popover as a string."""
        return self.browser.text(self._raw_title_el)

    @property
    def body(self):
        """Returns the text of the body of the popover as a string."""
        el = self.browser.element(self.BODY)
        return self.browser.text(el)

    @property
    def footer(self):
        """Returns the text of the body of the popover as a string."""
        el = self.browser.element(self.FOOTER)
        return self.browser.text(el)

    def close(self):
        """Close popover window."""
        self.browser.click(self.browser.element(self.CLOSE))


class Popover(BasePopover, Widget):
    ROOT = ParametrizedLocator("{@locator}")
    DEFAULT_LOCATOR = ".//div[contains(@role, 'dialog') and contains(@class, 'pf-c-popover')]"

    def __init__(self, parent, locator=None, logger=None):
        super().__init__(parent, logger=logger)
        if locator:
            self.locator = locator
        else:
            self.locator = self.DEFAULT_LOCATOR



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import DonutChart

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/charts/donut-chart"


@pytest.fixture
def donut_chart(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-donut-chart-right-aligned-legend']"
        donut_chart = DonutChart(locator="./div")

    return TestView(browser).donut_chart


def test_donut(donut_chart):
    assert donut_chart.donut.labels == ["100", "Pets"]
    assert donut_chart.legend.all_items == [
        {"label": "Cats", "value": "35"},
        {"label": "Dogs", "value": "55"},
        {"label": "Birds", "value": "10"},
    ]
    assert donut_chart.legend.item("Cats").label == "Cats"



from .select import Select


class BaseContextSelector:

    ITEMS_LOCATOR = ".//ul[@class='pf-c-context-selector__menu-list']/li"
    ITEM_LOCATOR = (
        ".//*[contains(@class, 'pf-c-context-selector__menu-list-item')"
        " and normalize-space(.)={}]"
    )
    SEARCH_INPUT_LOCATOR = ".//input[@type='search']"
    SEARCH_BUTTON_LOCATOR = ".//button[contains(@id, 'pf-context-selector-search-button')]"

    def item_select(self, item, use_search=False):
        """Opens the Context Selector and selects the desired item.

        Args:
            item: Item to be selected
            use_search: whether to search for item before selecting it
        """
        self.logger.info("Selecting %r in %r", item, self)
        with self.opened():
            if use_search:
                self.browser.send_keys(item, self.SEARCH_INPUT_LOCATOR)
                self.browser.click(self.SEARCH_BUTTON_LOCATOR)
            self.browser.click(self.item_element(item, close=False))


class ContextSelector(BaseContextSelector, Select):
    pass



import pytest
from wait_for import wait_for
from widgetastic.widget import View

from widgetastic_patternfly4 import PaginationNavDisabled
from widgetastic_patternfly4.ouia import Pagination

TESTING_PAGE_URL = (
    "https://patternfly-docs-ouia.netlify.app/documentation/react/components/pagination"  # noqa
)


@pytest.fixture
def paginator(browser, request):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-pagination-ouia']"
        paginator = Pagination("pagination-options-menu-top")

    paginator = TestView(browser).paginator
    wait_for(lambda: paginator.is_displayed, num_sec=10)
    yield paginator
    try:
        paginator.first_page()
    except PaginationNavDisabled:
        # We are already at the first page...
        pass


def test_first_page(paginator):
    paginator.last_page()
    assert paginator.current_page == 27
    paginator.first_page()
    assert paginator.is_first_disabled
    assert paginator.is_previous_disabled
    assert not paginator.is_next_disabled
    assert not paginator.is_last_disabled
    assert paginator.current_page == 1
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (1, 20)
    assert paginator.total_items == 523


def test_previous_page(paginator):
    paginator.next_page()
    assert paginator.current_page == 2
    paginator.previous_page()
    assert not paginator.is_next_disabled
    assert not paginator.is_last_disabled
    assert paginator.current_page == 1
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (1, 20)
    assert paginator.total_items == 523


def test_next_page(paginator):
    paginator.next_page()
    assert not paginator.is_first_disabled
    assert not paginator.is_previous_disabled
    assert not paginator.is_next_disabled
    assert not paginator.is_last_disabled
    assert paginator.current_page == 2
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (21, 40)
    assert paginator.total_items == 523


def test_last_page(paginator):
    paginator.last_page()
    assert not paginator.is_first_disabled
    assert not paginator.is_previous_disabled
    assert paginator.is_next_disabled
    assert paginator.is_last_disabled
    assert paginator.current_page == 27
    assert paginator.total_pages == 27
    assert paginator.displayed_items == (521, 523)
    assert paginator.total_items == 523


def test_per_page_options(paginator):
    assert paginator.per_page_options == [
        "10 per page",
        "20 per page",
        "50 per page",
        "100 per page",
    ]


@pytest.mark.parametrize("items_per_page", [50, 100])
def test_iteration(paginator, items_per_page):
    assert paginator.is_first_disabled
    paginator.set_per_page(items_per_page)
    assert paginator.current_per_page == items_per_page

    # Ensure we're always using an int for the math calculations
    items_per_page_int = int(str(items_per_page).split()[0])

    expected_total_pages = 523 // items_per_page_int + 1
    assert paginator.is_previous_disabled
    with paginator.cache_per_page_value():
        for page in paginator:
            assert paginator.current_page == page
            assert paginator.total_pages == expected_total_pages
            if items_per_page_int * page > paginator.total_items:
                right_number = paginator.total_items
            else:
                right_number = items_per_page_int * page
            assert paginator.displayed_items == (1 + items_per_page_int * (page - 1), right_number)
            assert paginator.total_items == 523
    assert paginator.is_next_disabled
    assert paginator.is_last_disabled


def test_bad_paginator_page_value(paginator):
    with pytest.raises(ValueError):
        paginator.set_per_page(9999999)
    with pytest.raises(ValueError):
        paginator.set_per_page("999999")


def test_custom_page(paginator):
    disp_items = paginator.displayed_items
    paginator.go_to_page(2)
    assert paginator.current_page == 2
    assert disp_items != paginator.displayed_items



from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Widget


class BaseProgress:
    """Represents the Patternfly Progress

    https://www.patternfly.org/v4/components/progress
    """

    DESCRIPTION = './/div[@class="pf-c-progress__description"]'
    MEASURE = './/span[@class="pf-c-progress__measure"]'
    PROGRESS_BAR = ".//div[@class='pf-c-progress__bar']"
    STATUS_TYPE_MAPPING = {
        "pf-m-warning": "warning",
        "pf-m-success": "success",
        "pf-m-danger": "error",
    }

    @property
    def description(self):
        el = self.browser.element(self.DESCRIPTION)
        return self.browser.text(el)

    @property
    def current_progress(self):
        return self.browser.element(self.PROGRESS_BAR).get_attribute("aria-valuenow")

    @property
    def status(self):
        """Returns the status of the progress, one of warning, success, error or info."""
        for class_ in self.browser.classes(self):
            if class_ in self.STATUS_TYPE_MAPPING:
                return self.STATUS_TYPE_MAPPING[class_]
        else:
            default_alert_type = "info"
            return default_alert_type


class Progress(BaseProgress, Widget):
    ROOT = ParametrizedLocator("{@locator}")

    def __init__(self, parent, locator, logger=None):
        super().__init__(parent, logger=logger)
        self.locator = locator



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import Card

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/card"


@pytest.fixture
def view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-card-basic']"
        card = Card("352")

    return TestView(browser)


def test_card_displayed(view):
    assert view.card.is_displayed



import pytest
from widgetastic.widget import ParametrizedView
from widgetastic.widget import View

from widgetastic_patternfly4 import CategoryChipGroup
from widgetastic_patternfly4 import Chip
from widgetastic_patternfly4 import ChipGroup
from widgetastic_patternfly4 import ChipReadOnlyError

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/chip-group"


@pytest.fixture(scope="module")
def chips_view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-chip-group-single']"
        chips = ParametrizedView.nested(Chip)

    return TestView(browser)


@pytest.fixture(scope="module")
def chip_group_view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-chip-group-simple-inline-chip-group']"

        non_existent_chip_group = ChipGroup(locator="foobar-locator")
        chip_group = ChipGroup()

    # Firefox fails the test if the chart is not fully visible therefore we click here on anchor
    # in order to properly scroll down
    anchor = browser.element("//a[@href='#simple-inline-chip-group']")
    browser.click(anchor)
    return TestView(browser)


@pytest.fixture(scope="module")
def category_chip_group_view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-chip-group-chip-groups-with-categories-removable']"
        category_one = CategoryChipGroup(label="Category one")
        category_two = CategoryChipGroup(label="Category two has a very long name")

    return TestView(browser)


def test_non_existent_chips(chip_group_view):
    assert not chip_group_view.non_existent_chip_group.is_displayed


def test_chipgroup_chips(chips_view):
    view = chips_view
    plain_chip = view.chips("Chip 1")
    long_chip = view.chips("Really long chip that goes on and on")
    chip_with_badge = view.chips("Chip")
    read_only_chip = view.chips("Read-only chip")

    assert plain_chip.text == "Chip 1"
    assert not plain_chip.badge
    assert plain_chip.is_displayed
    assert not plain_chip.read_only
    assert plain_chip.read() == "Chip 1"
    plain_chip.remove()
    assert not plain_chip.is_displayed

    assert long_chip.text == "Really long chip that goes on and on"
    assert not long_chip.badge
    assert long_chip.is_displayed
    assert not long_chip.read_only
    assert long_chip.read() == "Really long chip that goes on and on"
    long_chip.remove()
    assert not long_chip.is_displayed

    assert chip_with_badge.text == "Chip"
    assert chip_with_badge.badge == "7"
    assert chip_with_badge.is_displayed
    assert not chip_with_badge.read_only
    assert chip_with_badge.read() == "Chip"
    chip_with_badge.remove()
    assert not chip_with_badge.is_displayed

    assert read_only_chip.text == "Read-only chip"
    assert not read_only_chip.badge
    assert not read_only_chip.button.is_displayed
    assert read_only_chip.is_displayed
    assert read_only_chip.read_only
    assert read_only_chip.read() == "Read-only chip"
    with pytest.raises(ChipReadOnlyError):
        read_only_chip.remove()


def test_chipgroup_simple(chip_group_view):
    assert chip_group_view.is_displayed
    assert chip_group_view.chip_group.is_displayed

    chips = [
        "Chip one",
        "Really long chip that goes on and on",
        "Chip three",
        "Chip four",
        "Chip five",
    ]
    assert chip_group_view.chip_group.read() == chips

    chip_group_view.chip_group.show_less()
    chips = ["Chip one", "Really long chip that goes on and on", "Chip three"]
    assert [chip.text for chip in chip_group_view.chip_group.get_chips(show_more=False)] == chips

    chips = ["Chip one", "Chip three", "Chip four", "Chip five"]
    chip_group_view.chip_group.remove_chip_by_name("Really long chip that goes on and on")
    assert chip_group_view.chip_group.read() == chips

    chip_group_view.chip_group.remove_all_chips()
    assert not chip_group_view.chip_group.has_chips


def test_chipgroup_category(category_chip_group_view):
    assert category_chip_group_view.category_one.is_displayed
    assert category_chip_group_view.category_one.label == "Category one"

    category_chip_group_view.category_one.close()
    assert not category_chip_group_view.category_one.is_displayed

    # This tests that a category disappears after all chips are removed
    category_chip_group_view.category_two.remove_all_chips()
    assert not category_chip_group_view.category_two.is_displayed



from wait_for import wait_for
from widgetastic.widget import ParametrizedLocator
from widgetastic.widget import ParametrizedView
from widgetastic.widget import Text
from widgetastic.widget import View

from .button import Button


class ChipReadOnlyError(Exception):
    def __init__(self, chip, message):
        super().__init__(message)
        self.chip = chip


CHIP_ROOT = ".//div[@class='pf-c-chip' and not(contains(@class, 'pf-m-overflow'))]"
CHIP_TEXT = ".//span[contains(@class, 'pf-c-chip__text')]"
CHIP_BADGE = ".//span[contains(@class, 'pf-c-badge')]"
GROUP_ROOT = ".//div[contains(@class, 'pf-c-chip-group')]"
CATEGORY_GROUP_ROOT = (
    ".//div[contains(@class, 'pf-c-chip-group') and contains(@class, 'pf-m-category')]"
)
CATEGORY_LABEL = ".//span[contains(@class, 'pf-c-chip-group__label')]"
CATEGORY_CLOSE = "./div[contains(@class, 'pf-c-chip-group__close')]/button"
# For backwards compatibility
OLD_GROUP_ROOT = ".//ul[contains(@class, 'pf-c-chip-group')]"
TOOLBAR_GROUP_LABEL = "./li/*[contains(@class, 'pf-c-chip-group__label')]"
STANDALONE_GROUP_LABEL = "./preceding-sibling::*[contains(@class, 'pf-c-chip-group__label')]"
OLD_CHIP_ROOT = (
    ".//*[(self::div or self::li) and contains(@class, 'pf-c-chip') "
    "and not(contains(@class, 'pf-m-overflow'))]"
)


class _BaseChip(View):
    """
    Holds attributes shared by both Chip and OverflowChip
    """

    _text = Text(CHIP_TEXT)
    _badge = Text(f"{CHIP_TEXT}/{CHIP_BADGE}")
    button = Button(**{"aria-label": "close"})

    @property
    def badge(self):
        """
        Return the text of the badge displayed on the chip, if it has a badge
        """
        return self._badge.text if self._badge.is_displayed else None

    @property
    def text(self):
        """
        Return the text displayed on the chip
        """
        if self.badge:
            # If this chip has a badge, strip the badge off the end to return only the text
            return self._text.text.rstrip(self.badge)
        return self._text.text

    @property
    def is_displayed(self):
        """Returns True if the component is displayed."""
        return self._text.is_displayed

    def read(self):
        """Returns the text of the Chip as a string"""
        return self.text


class Chip(ParametrizedView, _BaseChip):
    PARAMETERS = ("text",)
    ROOT = ParametrizedLocator(
        f"{CHIP_ROOT}[{CHIP_TEXT}[starts-with(normalize-space(.), {{text|quote}})]]"
    )

    @staticmethod
    def _get_text_ignoring_badge(browser, element):
        el = element
        badge_elements = browser.elements(CHIP_BADGE, parent=el)
        if badge_elements:
            badge = browser.text(el)
            return badge.rstrip(browser.text(badge_elements[0]))
        return browser.text(el)

    @classmethod
    def all(cls, browser):
        """Returns a list of the text of each chip"""
        return [
            (cls._get_text_ignoring_badge(browser, el),)
            for el in browser.elements(f"{CHIP_ROOT}/{CHIP_TEXT}")
        ]

    def __init__(self, *args, **kwargs):
        ParametrizedView.__init__(self, *args, **kwargs)

    def remove(self):
        """Removes a chip"""

        def _gone():
            return not self.is_displayed

        if not self.read_only:
            self.button.click()
            wait_for(_gone, num_sec=3, message="wait for chip to dissappear", delay=0.1)
        else:
            raise ChipReadOnlyError(self, "Chip is read-only")

    @property
    def read_only(self):
        """Returns a boolean detailing if the chip is read only"""
        return not self.button.is_displayed


class OverflowChip(_BaseChip):
    """
    The 'Show More'/'Show Less' button is essentially a special kind of chip
    """

    ROOT = ".//button[contains(@class, 'pf-c-chip') and contains(@class, 'pf-m-overflow')]"

    def _show_less_shown(self):
        return self.text.replace(" ", "").lower() == "showless"

    def _show_more_shown(self):
        return "more" in self.text.lower()

    def show_more(self):
        """Expands a chip group"""
        if self._show_more_shown():
            self._text.click()
        wait_for(
            func=self._show_less_shown,
            num_sec=3,
            delay=0.1,
            message="wait for 'show less' button to appear",
        )

    def show_less(self):
        """Collapses a chip group"""
        if self._show_less_shown():
            self._text.click()
        wait_for(
            self._show_more_shown,
            num_sec=3,
            delay=0.1,
            message="wait for 'show more' button to appear",
        )


class ChipGroup(View):
    """
    Represents a chip group that is "on its own", i.e. not a part of a chip group category
    """

    ROOT = ParametrizedLocator("{@locator}")

    overflow = OverflowChip()
    chips = ParametrizedView.nested(Chip)

    def __init__(self, parent, locator=None, logger=None):
        super().__init__(parent, logger=logger)
        self.locator = locator or GROUP_ROOT

    # For backwards compatibility
    @property
    def label(self):
        # It's unlikely we'll have a labelled chip group that is not in a toolbar
        # ... but just in case
        elements = self.browser.elements(STANDALONE_GROUP_LABEL)
        return self.browser.text(elements[0]) if elements else None

    # For backwards compatibility
    @property
    def is_multiselect(self):
        return self.overflow.is_displayed

    def show_more(self):
        """Expands a chip group"""
        self.overflow.show_more()

    def show_less(self):
        """Collapses a chip group"""
        self.overflow.show_less()

    def get_chips(self, show_more=True):
        """
        A helper to expand the chip group before reading its chips
        """
        if self.overflow.is_displayed and show_more:
            self.show_more()
        return self.chips

    @property
    def has_chips(self):
        return self.is_displayed

    def __iter__(self):
        chips = self.get_chips()
        for chip in chips:
            yield chip

    def remove_chip_by_name(self, name):
        """Removes a chip from the group by name"""
        for chip in self:
            if chip.text.lower() == name.lower():
                chip.remove()
                break
        else:
            raise ValueError(f"Could not find chip with name '{name}'")

    def remove_all_chips(self):
        """Removes all chips from the group"""
        for chip in self:
            chip.remove()

    def read(self):
        self.logger.info(self.get_chips())
        return [chip.text for chip in self]


class CategoryChipGroup(ChipGroup):
    """
    Represents a Chip Group with a category label
    """

    ROOT = ParametrizedLocator(
        f"{CATEGORY_GROUP_ROOT}[{CATEGORY_LABEL}[normalize-space(.)={{@_label|quote}}]]"
    )

    chips = ParametrizedView.nested(Chip)
    close_button = Button(locator=CATEGORY_CLOSE)

    def __init__(self, parent, label, logger=None):
        self._label = label
        ChipGroup.__init__(self, parent, logger=logger)

    @property
    def label(self):
        elements = self.browser.elements(CATEGORY_LABEL)
        return self.browser.text(elements[0]) if elements else None

    @property
    def can_close(self):
        return self.close_button.is_displayed

    def close(self):
        self.close_button.click()

    @classmethod
    def all(cls, browser):
        """Returns a list of all category labels."""
        return [browser.text(el) for el in browser.elements(CATEGORY_LABEL)]


# For backwards compatibility
class OldChip(Chip):
    ROOT = ParametrizedLocator(
        f"{OLD_CHIP_ROOT}[{CHIP_TEXT}[starts-with(normalize-space(.), {{text|quote}})]]"
    )

    @classmethod
    def all(cls, browser):
        """Returns a list of the text of each chip"""
        return [
            (cls._get_text_ignoring_badge(browser, el),)
            for el in browser.elements(f"{OLD_CHIP_ROOT}/{CHIP_TEXT}")
        ]


# For backwards compatibility
class OldOverflowChip(OverflowChip):
    ROOT = (
        ".//*[(self::li or self::div) and "
        "contains(@class, 'pf-c-chip') and contains(@class, 'pf-m-overflow')]"
    )

    @property
    def is_displayed(self):
        """Returns a boolean detailing if the overflow chip is displayed"""
        return self.button.is_displayed


# For backwards compatibility
class StandAloneChipGroup(ChipGroup):
    chips = ParametrizedView.nested(OldChip)
    overflow = ParametrizedView.nested(OldOverflowChip)


# For backwards compatibility
class ChipGroupToolbarCategory(ParametrizedView, ChipGroup):
    """
    Represents a chip group that is part of a toolbar, identifiable by a category label
    """

    PARAMETERS = ("label",)
    ROOT = ParametrizedLocator(
        f"{OLD_GROUP_ROOT}[{TOOLBAR_GROUP_LABEL}[normalize-space(.)={{label|quote}}]]"
    )

    chips = ParametrizedView.nested(OldChip)

    def __init__(self, *args, **kwargs):
        ParametrizedView.__init__(self, *args, **kwargs)

    @property
    def label(self):
        elements = self.browser.elements(CATEGORY_LABEL)
        return self.browser.text(elements[0]) if elements else None

    @classmethod
    def all(cls, browser):
        return [
            (browser.text(el),)
            for el in browser.elements(f"{OLD_GROUP_ROOT}/{TOOLBAR_GROUP_LABEL}")
        ]


# For backwards compatibility
class ChipGroupToolbar(View):
    ROOT = ParametrizedLocator("{@locator}")

    # The parent of the chip group toolbar can be any element type
    # The locator should be the parent node which holds all the pf-c-chip-group elements
    TOOLBAR_LOCATOR = (
        ".//ul[contains(@class, 'pf-c-chip-group') and "
        "contains(@class, 'pf-m-toolbar')]/parent::*"
    )

    overflow = OldOverflowChip(
        locator=("./div[contains(@class, 'pf-c-chip') and contains(@class, 'pf-m-overflow')]")
    )
    groups = ParametrizedView.nested(ChipGroupToolbarCategory)

    def __init__(self, parent, locator=None, logger=None):
        self.locator = locator or self.TOOLBAR_LOCATOR
        super().__init__(parent, logger=logger)

    def get_groups(self, show_more=True):
        """
        A helper to expand the chip group toolbar before reading its groups
        """
        if self.overflow.is_displayed and show_more:
            self.overflow.show_more()
        return self.groups

    def __iter__(self):
        for group in self.get_groups():
            yield group

    def read(self):
        """Returns a dict of chips"""
        groups = {}
        for group in self:
            groups[group.label] = group.read()
        return groups

    def show_more(self):
        """Expands a chip group"""
        self.overflow.show_more()

    def show_less(self):
        """Collapses a chip group"""
        self.overflow.show_less()

    @property
    def has_chips(self):
        # If we delete all chips the ROOT is still shown thus we need to check if there are
        # any chips.
        return self.read() != {}



from collections import OrderedDict

from wait_for import wait_for
from widgetastic.exceptions import NoSuchElementException
from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Widget
from widgetastic.xpath import quote


def check_nav_loaded(fn):
    def inner(self, *args, **kwargs):
        assert self.loaded
        return fn(self, *args, **kwargs)

    return inner


class NavSelectionNotFound(Exception):
    pass


class BaseNavigation:
    """The Patternfly navigation.

    https://www.patternfly.org/v4/documentation/react/components/nav
    """

    CURRENTLY_SELECTED = (
        './/*[self::a or self::button][contains(@class, "pf-m-current") or '
        'parent::li[contains(@class, "pf-m-current")]]'
    )
    ITEMS = "./ul/li/*[self::a or self::button]"
    SUB_ITEMS_ROOT = "./section"
    ITEM_MATCHING = "./ul/li[.//*[self::a or self::button][normalize-space(.)={}]]"

    @property
    def loaded(self):
        """Returns a boolean detailing if the nav is loaded."""
        out = self.browser.element(".").get_attribute("data-ouia-safe")
        if out == "false":
            self.logger.info("Navigation not ready yet")
            wait_for(
                lambda: self.browser.element(".").get_attribute("data-ouia-safe") == "true",
                num_sec=10,
            )
        elif not out:
            self.logger.info("Navigation doesn't have 'data-ouia-safe' property")
        return True

    @check_nav_loaded
    def read(self):
        """Returns the current navigation item."""
        return self.currently_selected

    @check_nav_loaded
    def nav_links(self, *levels):
        """Returns a list of all navigation items."""
        if not levels:
            return [
                el.get_property("textContent").strip() for el in self.browser.elements(self.ITEMS)
            ]
        current_item = self
        for i, level in enumerate(levels):
            li = self.browser.element(
                self.ITEM_MATCHING.format(quote(level.strip())), parent=current_item
            )

            try:
                current_item = self.browser.element(self.SUB_ITEMS_ROOT, parent=li)
            except NoSuchElementException:
                if i == len(levels) - 1:
                    return []
                else:
                    raise
        return [
            el.get_property("textContent").strip()
            for el in self.browser.elements(self.ITEMS, parent=current_item)
        ]

    @check_nav_loaded
    def nav_item_tree(self, start=None):
        """Returns an ordered dict representing the navigation tree."""
        start = start or []
        result = OrderedDict()
        for item in self.nav_links(*start):
            sub_items = self.nav_item_tree(start=start + [item])
            result[item] = sub_items or None
        if result and all(value is None for value in result.values()):
            # If there are no child nodes, then just make it a list
            result = list(result.keys())
        return result

    @property
    @check_nav_loaded
    def currently_selected(self):
        """Returns the currently selected item."""
        return [self.browser.text(el) for el in self.browser.elements(self.CURRENTLY_SELECTED)]

    @check_nav_loaded
    def select(self, *levels, **kwargs):
        """Select an item in the navigation.

        Args:
            *levels: Items to be clicked in the navigation.
            force: Force navigation to happen, defaults to False.
        """
        self.logger.info("Selecting %r in navigation", levels)
        force = kwargs.get("force", False)
        if not force and list(levels) == self.currently_selected:
            return
        current_item = self
        for i, level in enumerate(levels, 1):
            try:
                li = self.browser.element(
                    self.ITEM_MATCHING.format(quote(level)), parent=current_item
                )
            except NoSuchElementException:
                raise NavSelectionNotFound(
                    f"{levels} not found in navigation tree. "
                    f"Could not find element: '{self.ITEM_MATCHING.format(quote(level))}'"
                )
            if "pf-m-expanded" not in li.get_attribute("class").split():
                self.browser.click(li)
            if i == len(levels):
                return
            current_item = self.browser.element(self.SUB_ITEMS_ROOT, parent=li)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.ROOT)


class Navigation(BaseNavigation, Widget):
    ROOT = ParametrizedLocator("{@locator}")
    LOCATOR_START = './/nav[@class="pf-c-nav"{}]'

    def __init__(self, parent, label=None, id=None, locator=None, logger=None):
        super().__init__(parent, logger=logger)

        quoted_label = quote(label) if label else ""
        if label:
            label_part = " and @label={} or @aria-label={}".format(quoted_label, quoted_label)
        else:
            label_part = ""

        id_part = " and @id={}".format(quote(id)) if id else ""
        if locator is not None:
            self.locator = locator
        elif label_part or id_part:
            self.locator = self.LOCATOR_START.format(label_part + id_part)
        else:
            raise TypeError("You need to specify either, id, label or locator for Navigation")



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Button

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/button"


@pytest.fixture
def variations_view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-button-variations']"
        any_button = Button()
        button1 = Button("Primary")

    return TestView(browser)


@pytest.fixture
def disabled_view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-button-disabled']"
        button = Button("Primary disabled")

    return TestView(browser)


def test_button_click(variations_view):
    assert variations_view.any_button.is_displayed
    assert variations_view.button1.is_displayed


def test_disabled_button(disabled_view):
    assert disabled_view.button.is_displayed
    assert disabled_view.button.disabled



from widgetastic.exceptions import NoSuchElementException

from .dropdown import Dropdown
from .dropdown import DropdownItemDisabled
from .dropdown import DropdownItemNotFound


class MenuItemDisabled(DropdownItemDisabled):
    pass


class MenuItemNotFound(DropdownItemNotFound):
    pass


class BaseMenu:
    """Represents the Patternfly Menu.

    Most menu's exist without a toggle to open them (e.g. all the examples in the link below).
    That is why there is the class level constant IS_ALWAYS_OPEN which defaults to 'True'.

    However, it is possible to have a menu that will only appear based on a button or toggle.
    In these cases you should inherit from menu and set IS_ALWAYS_OPEN to 'False'. BUTTON_LOCATOR
    will also need to be modified to point to the toggle button. In the DOM, the menu widget is
    usually placed as a sibling to the toggle button. You can either override ROOT or simply pass
    a locator when the widget is instantiated to point to the parent of the toggle button.

    .. code-block:: python

        class PermissionsMenu(Menu):
            IS_ALWAYS_OPEN = False
            BUTTON_LOCATOR = ".//button[contains(@class, 'pf-c-menu-toggle')]"
            ROOT = f"{BUTTON_LOCATOR}/.."

    https://www.patternfly.org/v4/documentation/react/components/menu
    """

    # most menus are always open
    IS_ALWAYS_OPEN = True

    BUTTON_LOCATOR = "./button"
    ITEMS_LOCATOR = ".//li[contains(@class, 'pf-c-menu__list-item')]"
    SELECTED_ITEMS_LOCATOR = ".//button[contains(@class, 'pf-m-selected')]"
    ITEM_LOCATOR = ".//*[contains(@class, 'pf-c-menu__list-item') and normalize-space(.)={}]"
    TEXT_LOCATOR = ".//div[contains(@class, 'pf-c-menu') and child::button[normalize-space(.)={}]]"

    @property
    def selected_items(self):
        """Returns a list of all selected items as strings."""
        with self.opened():
            result = [
                self.browser.text(el) for el in self.browser.elements(self.SELECTED_ITEMS_LOCATOR)
            ]
        return result

    @property
    def is_open(self):
        """
        Returns True if the Dropdown is open

        Some menus are visible without having to click anything, so if the items are displayed
        we consider that to be open as well.
        """
        return self.IS_ALWAYS_OPEN or "pf-m-expanded" in self.browser.classes(self.BUTTON_LOCATOR)

    @property
    def is_enabled(self):
        """Returns True if the menu toggle itself is enabled and therefore interactive."""
        return self.IS_ALWAYS_OPEN or "disabled" not in self.browser.classes(self.BUTTON_LOCATOR)

    def close(self, ignore_nonpresent=False):
        """Close the menu

        It it is always open we do nothing

        Args:
            ignore_nonpresent: Will ignore exceptions due to disabled or missing dropdown
        """
        try:
            self._verify_enabled()
            if self.IS_ALWAYS_OPEN:
                self.logger.info("Tried to close %r but it's always open. Ignoring.", self)
                return
            if self.is_open:
                self.browser.click(self.BUTTON_LOCATOR)
        except (NoSuchElementException, MenuItemDisabled):
            if ignore_nonpresent:
                self.logger.info("%r hid so it was not possible to close it. But ignoring.", self)
            else:
                raise

    def item_element(self, item, close=True):
        """Returns a WebElement for given item name."""
        try:
            return super().item_element(item, close)
        except DropdownItemNotFound:
            raise MenuItemNotFound(
                "Item {!r} not found in {}. Available items: {}".format(
                    item, repr(self), self.items
                )
            )

    def item_select(self, item):
        """Opens the Menu and selects the desired item.

        Args:
            item: Item to be selected

        Raises:
            MenuItemDisabled: if item is disabled
        """
        try:
            return super().item_select(item)
        except DropdownItemDisabled:
            raise MenuItemDisabled(
                'Item "{}" of {} is disabled\n'
                "The following items are available and enabled: {}".format(
                    item, repr(self), self.enabled_items
                )
            )

    def fill(self, value):
        """Fills a Menu with a value or values."""
        if isinstance(value, list):
            for val in value:
                self.item_select(val)
        else:
            self.item_select(value)

    def read(self):
        """Returns a string of the text of the selected option."""
        if self.selected_items:
            return self.browser.text(self.SELECTED_ITEMS_LOCATOR)
        else:
            return self.browser.text(self)


class Menu(BaseMenu, Dropdown):
    DEFAULT_LOCATOR = './/div[contains(@class, "pf-c-menu")]'


class BaseCheckboxMenu(BaseMenu):
    """
    Represents a checkbox menu.
    """

    ITEM_LOCATOR = ".//label[normalize-space(.)={}]/preceding-sibling::input"  # noqa

    def item_select(self, items, close=True):
        """Opens the Checkbox and selects the desired item.

        Args:
            item: Item to be selected
            close: Close the dropdown when finished
        """
        if not isinstance(items, (list, tuple, set)):
            items = [items]

        try:
            for item in items:
                element = self.item_element(item, close=False)
                if not self.browser.is_selected(element):
                    element.click()
        finally:
            if close:
                self.close()

    def item_deselect(self, items, close=True):
        """Opens the Checkbox and deselects the desired item.

        Args:
            item: Item to be selected
            close: Close the dropdown when finished
        """
        if not isinstance(items, (list, tuple, set)):
            items = [items]

        try:
            for item in items:
                element = self.item_element(item, close=False)
                if self.browser.is_selected(element):
                    element.click()
        finally:
            if close:
                self.close()

    def fill(self, items):
        """Fills a Checkbox with all items.
        Example dictionary: {"foo": True, "bar": False, "baz": True}

        Args:
            items: A dictionary containing what items to select (True) or deselect (False)
        """
        try:
            for item, value in items.items():
                if value:
                    self.item_select(item, close=False)
                else:
                    self.item_deselect(item, close=False)
        finally:
            self.close()

    def read(self):
        """Returns a dictionary containing the selected status as bools."""
        selected = {}
        with self.opened():
            for el in self.browser.elements(self.ITEMS_LOCATOR):
                item = self.browser.text(el)
                try:
                    # get the child element of the label
                    selected[item] = self.browser.element(
                        parent=el, locator="./input"
                    ).is_selected()
                except NoSuchElementException:
                    selected[item] = False

        return selected

    def _get_items(self, close=False):
        """Returns a list of all checkbox items as strings.

        Args:
            close: Close the dropdown when finished
        """
        self.open()
        result = [self.browser.text(el) for el in self.browser.elements(self.ITEMS_LOCATOR)]

        if close:
            self.close()

        return result

    @property
    def items(self):
        """Returns a list of all CheckboxSelect items as strings."""
        return self._get_items(close=True)


class CheckboxMenu(BaseCheckboxMenu, Dropdown):
    DEFAULT_LOCATOR = './/div[contains(@class, "pf-c-menu")]'



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import FormSelect
from widgetastic_patternfly4 import FormSelectDisabled
from widgetastic_patternfly4 import FormSelectOptionDisabled
from widgetastic_patternfly4 import FormSelectOptionNotFound

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/form-select"


class FormSelectTestView(View):
    ROOT = ".//main"

    input = FormSelect(locator=".//div[@id='ws-react-c-form-select-basic']/select")
    input_grouping = FormSelect(locator=".//div[@id='ws-react-c-form-select-grouped']/select")
    input_invalid = FormSelect(locator=".//div[@id='ws-react-c-form-select-invalid']/select")
    input_disabled = FormSelect(locator=".//div[@id='ws-react-c-form-select-disabled']/select")


@pytest.fixture
def view(browser):
    return FormSelectTestView(browser)


def test_formselect_visibility(view):
    assert view.input.is_displayed
    assert view.input_grouping.is_displayed
    assert view.input_invalid.is_displayed
    assert view.input_disabled.is_displayed


def test_formselect_enablement(view):
    assert view.input.is_enabled
    assert view.input_grouping.is_enabled
    assert view.input_invalid.is_enabled
    assert not view.input_disabled.is_enabled


def test_formselect_validity(view):
    assert view.input.is_valid
    assert view.input_grouping.is_valid
    assert view.input_disabled.is_valid
    assert view.input_invalid.is_valid
    view.input_invalid.fill("One")
    assert view.input_invalid.is_valid
    view.input_invalid.fill("Select a number")
    assert not view.input_invalid.is_valid


def test_formselect_value(view):
    assert len(view.input.all_options) == 7
    assert len(view.input.all_enabled_options) == 6
    assert "Mrs" in view.input.all_options
    view.input.fill("Mrs")
    assert view.input.read() == "Mrs"


def test_formselect_option_enablement(view):
    expected_enabled_options = {"Mr", "Miss", "Mrs", "Ms", "Dr", "Other"}
    expected_disabled_options = {"Please Choose"}
    assert set(view.input.all_enabled_options) == expected_enabled_options
    assert expected_disabled_options not in set(view.input.all_enabled_options)
    expected_enabled_optgroup_options = {
        "The first option",
        "Second option is selected by default",
        "The third option",
        "The fourth option",
    }
    expected_disabled_optgroup_options = {"The fifth option", "The six option"}
    assert set(view.input_grouping.all_enabled_options) == expected_enabled_optgroup_options
    assert expected_disabled_optgroup_options not in set(view.input_grouping.all_enabled_options)


def test_formselect_fill_disabled_select(view):
    with pytest.raises(FormSelectDisabled):
        view.input_disabled.fill("Two")


def test_formselect_fill_disabled_option(view):
    with pytest.raises(FormSelectOptionDisabled):
        view.input_grouping.fill("The fifth option")


def test_formselect_fill_nonexistent_option(view):
    with pytest.raises(FormSelectOptionNotFound):
        view.input.fill("foo")



from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import ClickableMixin
from widgetastic.widget import Widget
from widgetastic.xpath import quote


class BaseButton:
    CHECK_VISIBILITY = True

    # Classes usable in the constructor
    # Button types
    PRIMARY = "pf-m-primary"
    SECONDARY = "pf-m-secondary"
    TETRIARY = "pf-m-tertiary"
    DANGER = "pf-m-danger"
    LINK = "pf-m-link"
    PLAIN = "pf-m-plain"

    # Shape
    BLOCK = "pf-m-block"

    def read(self):
        """Returns the string of the button."""
        return self.browser.text(self)

    @property
    def active(self):
        """Returns a boolean detailing if the button is active."""
        return "pf-m-active" in self.browser.classes(self)

    @property
    def disabled(self):
        """Returns a boolean detailing if the button is disabled."""
        check1 = "pf-m-disabled" in self.browser.classes(self)
        check2 = "pf-m-aria-disabled" in self.browser.classes(self)
        return check1 or check2 or self.browser.get_attribute("disabled", self) is not None

    def __repr__(self):
        return "{}{}".format(type(self).__name__, self.locator)

    @property
    def title(self):
        """Returns the title of the button as a string."""
        return self.browser.get_attribute("title", self)


class Button(BaseButton, Widget, ClickableMixin):
    """A Patternfly button

    You can match by text, partial text or by attributes, you can also add the patternfly classes
    into the matching.

    .. code-block:: python

        Button("Text of button (unless it is an input ...)")
        Button("contains", "Text of button (unless it is an input ...)")
        Button(title="Show xyz")  # And such
        Button("Add", classes=[Button.PRIMARY])
        Button(locator=".//xpath")
        assert button.active
        assert not button.disabled
    """

    ROOT = ParametrizedLocator("{@locator}")

    def _generate_locator(self, *text, **kwargs):
        classes = kwargs.pop("classes", [])
        if text:
            if kwargs:  # classes should have been the only kwarg combined with text args
                raise TypeError("If you pass button text then only pass classes in addition")
            if len(text) == 1:
                locator_conditions = "normalize-space(.)={}".format(quote(text[0]))
            elif len(text) == 2 and text[0].lower() == "contains":
                locator_conditions = "contains(normalize-space(.), {})".format(quote(text[1]))
            else:
                raise TypeError("An illegal combination of args/kwargs")
        else:
            # Join the kwargs, if any
            locator_conditions = " and ".join(
                "@{}={}".format(attr, quote(value)) for attr, value in kwargs.items()
            )

        if classes:
            if locator_conditions:
                locator_conditions += " and "
            locator_conditions += " and ".join(
                "contains(@class, {})".format(quote(klass)) for klass in classes
            )
        if locator_conditions:
            locator_conditions = "and ({})".format(locator_conditions)

        return (
            ".//*[(self::a or self::button or (self::input and "
            "(@type='button' or @type='submit'))) and "
            f"contains(@class, 'pf-c-button') {locator_conditions}]"
        )

    def __init__(self, parent, *text, **kwargs):
        super().__init__(parent, logger=kwargs.pop("logger", None))
        self.args = text
        self.kwargs = kwargs
        self.locator = kwargs.pop("locator", self._generate_locator(*text, **kwargs))



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import PatternflyTable

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/table"

SORT = [
    ("Repositories", "ascending", ["a", "one", "p"]),
    ("Repositories", "descending", ["p", "one", "a"]),
    ("Pull requests", "ascending", ["a", "b", "k"]),
    ("Pull requests", "descending", ["k", "b", "a"]),
]


@pytest.mark.parametrize("sample", SORT, ids=lambda sample: "{}-{}".format(sample[0], sample[1]))
def test_sortable_table(browser, sample):
    header, order, expected_result = sample

    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-table-ouia']"
        table = PatternflyTable("Sortable Table")

    view = TestView(browser)
    view.table.sort_by(header, order)
    column = [row[header] for row in view.table.read()]
    assert column == expected_result



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import Navigation

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/nav"


@pytest.fixture
def view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-nav-ouia']"
        nav = Navigation("Nav Default")

    return TestView(browser)


def test_navigation(browser, view):
    assert view.nav.currently_selected == ["Link 1"]
    assert view.nav.nav_item_tree() == ["Link 1", "Link 2", "Link 3", "Link 4"]



import six
from selenium.common.exceptions import NoSuchElementException
from widgetastic.log import create_item_logger
from widgetastic.widget import Table
from widgetastic.widget import TableColumn
from widgetastic.widget import TableRow
from widgetastic.widget import Text
from widgetastic.widget import Widget
from widgetastic.widget.table import resolve_table_widget


class HeaderColumn(TableColumn):
    """Represents a cell in the header row."""

    def __locator__(self):
        return "(./td|./th)[{}]".format(self.position + 1)

    @property
    def is_sortable(self):
        """Returns true of the column is sortable."""
        return "pf-c-table__sort" in self.browser.classes(self)

    @property
    def sorting_order(self):
        """Returns current sorting order as a string."""
        return self.browser.get_attribute("aria-sort", self)

    def sort(self, order="ascending"):
        """Sorts the column according to the supplied "ascending" or "descending"."""
        if order not in ("ascending", "descending"):
            raise ValueError("order should be either 'ascending' or 'descending'")
        for _ in range(10):
            if self.sorting_order == order:
                break
            self.click()
        else:
            raise AssertionError(f"Could not set sorting order to `{order}` after 10 tries.")


class HeaderRow(TableRow):
    Column = HeaderColumn

    def __init__(self, parent, logger=None):
        Widget.__init__(self, parent, logger=logger)

    def __locator__(self):
        return "./thead/tr"

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.parent)

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            index = self.table.header_index_mapping[self.table.ensure_normal(item)]
        elif isinstance(item, int):
            index = item
        else:
            raise TypeError("row[] accepts only integers and strings")
        return self.Column(self, index, logger=create_item_logger(self.logger, item))

    def read(self):
        """Returns the values of the headers of the HeaderRow object."""
        return self.parent.headers


class PatternflyTableRow(TableRow):
    """
    Extends TableRow to support having a 'th' tag within the row
    """

    HEADER_IN_ROW = "./th[1]"
    TABLE_COLUMN_CLS = TableColumn

    @property
    def has_row_header(self):
        """Returns a boolean detailing if the Table Row has a header."""
        return len(self.browser.elements(self.HEADER_IN_ROW)) > 0

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            index = self.table.header_index_mapping[self.table.ensure_normal(item)]
        elif isinstance(item, int):
            index = item
        else:
            raise TypeError("row[] accepts only integers and strings")

        # We need to do adjustments if a <th> tag exists inside the row...
        # Typically the layout is: <td>, <th>, <td>, <td>, <td>, and so on...
        if self.has_row_header:
            if index == 1:
                # We assume the header entry always sits at position 1. Return a TableColumn for it.
                # Pass position '0' since the Column __locator__ uses 'position + 1'
                return self.TABLE_COLUMN_CLS(self, 0, logger=create_item_logger(self.logger, item))

            if index > 1:
                # Adjust the index for td objects that exist beyond the th so xpath is valid
                index = index - 1
        # After adjusting the index, call the original __getitem__ to get our TableColumn item
        return super(PatternflyTableRow, self).__getitem__(index)


class BasePatternflyTable:
    """Represents the Patternfly table.

    https://www.patternfly.org/v4/documentation/react/components/table
    """

    ROWS = "./tbody/tr[./td]"
    HEADERS = "./thead/tr/th|./tr/th|./thead/tr/td"

    Row = PatternflyTableRow

    header_row = HeaderRow()

    @property
    def _is_header_in_body(self):
        """Override this to return False.

        Some PF4 tables have a 'header cell' in the row, which is a th in the row, this will
        cause Table._is_header_in_body to incorrectly return 'True'
        """
        return False

    def sort_by(self, column, order):
        """Sets the sort order for the supplied column by name, and "ascending/descending"."""
        header = self.header_row[column]
        header.sort(order)

    def _toggle_select_all(self, value, column):
        header = self.header_row[column]
        header.fill(value)

    def select_all(self, column=0):
        """Selects all the rows."""
        self._toggle_select_all(True, column)

    def deselect_all(self, column=0):
        """Deselects all the rows."""
        self._toggle_select_all(False, column)


class PatternflyTable(BasePatternflyTable, Table):
    pass


class ExpandableTableHeaderColumn(TableColumn):
    """
    Used for special cases where a <th> appears as a column in ExpandableTable.
    """

    def __locator__(self):
        """Override the locator to look inside the first 'tr' within the tbody"""
        return "./tr[1]/th[{}]".format(self.position + 1)


class RowNotExpandable(Exception):
    def __init__(self, row):
        self.row = row

    def __str__(self):
        return "Row is not expandable: {}".format(repr(self.row))


class ColumnNotExpandable(Exception):
    def __init__(self, column):
        self.column = column

    def __str__(self):
        return "Column is not expandable: {}".format(repr(self.column))


class ExpandableTableRow(PatternflyTableRow):
    """Represents a row in the table.

    If subclassing and also changing the Column class, do not forget to set the Column to the new
    class.

    Args:
        index: Position of the row in the table.
    """

    ROW = "./tr[1]"
    EXPANDABLE_CONTENT = "./tr[2]"

    # Override these values inherited from PatternflyTableRow...
    HEADER_IN_ROW = "./tr[1]/th[1]"
    TABLE_COLUMN_CLS = ExpandableTableHeaderColumn

    def __init__(self, parent, index, content_view=None, logger=None):
        super(ExpandableTableRow, self).__init__(parent, index, logger=logger)

        content_parent = Text(parent=self, locator=self.EXPANDABLE_CONTENT)
        if content_view:
            self.content = resolve_table_widget(content_parent, content_view)
        else:
            self.content = content_parent

    @property
    def is_displayed(self):
        """Returns a boolean detailing if the Table Row is displayed."""
        return self.browser.is_displayed(locator=self.ROW)

    @property
    def is_expandable(self):
        """Returns a boolean detailing if the table row is expandable."""
        return self[0].widget.is_displayed

    def _check_expandable(self):
        if not self.is_expandable:
            raise RowNotExpandable(self)

    @property
    def is_expanded(self):
        """Returns a boolean detailing if the table row has been expanded."""
        self._check_expandable()
        return self.browser.is_displayed(locator=self.EXPANDABLE_CONTENT)

    def expand(self):
        """Expands the table row."""
        self._check_expandable()
        if not self.is_expanded:
            self[0].widget.click()
            self.content.wait_displayed()

    def collapse(self):
        """Collapses the table row."""
        self._check_expandable()
        if self.is_expanded:
            self[0].widget.click()

    def read(self):
        """Returns a text representation of the table row."""
        result = super(ExpandableTableRow, self).read()
        # Remove the column with the "expand" button in it
        if 0 in result and not result[0]:
            del result[0]
        return result


class BaseExpandableTable:
    """
    The patternfly 4 expandable table has the following outline:

    .. code-block:: html

        <table>
          <thead>
          <tbody>
            <tr>The row always on display.</tr>
            <tr>The "expandable" content viewed by clicking the arrow button</tr>
          </tbody>
          <tbody>
            <tr>Next row...</tr>
            <tr>Next row's expandable content...</tr>

    Therefore, we modify the behavior of Table here to look for rows based on 'tbody'
    tags instead of 'tr' tags. We use a custom class, ExpandableTableRow, which treats
    the first <tr> tag as a normal table row (if you call row.read(), it will read the
    this row -- also table column widgets will apply to this row, etc. etc.), but it
    will treat the second <tr> tag as a Text widget, or a parent for a custom defined View
    """

    ROWS = "./tbody"
    ROW_RESOLVER_PATH = "/table/tbody"
    ROW_AT_INDEX = "./tbody[{0}]"
    COLUMN_RESOLVER_PATH = "/tr[0]/td"
    COLUMN_AT_POSITION = "./tr[1]/td[{0}]"
    ROW_TAG = "tbody"
    HEADERS = "./thead/tr/th|./thead/tr/td"
    Row = ExpandableTableRow

    def __init__(self, *args, **kwargs):
        """Extend init of Table

        Automatically add the 'expand' button widget as column 0.

        Provide additional kwarg for 'content_view', which is used to pass in a WidgetDescriptor
        to be used as the Widget/View for the expanded content of each row.
        """
        column_widgets = kwargs.get("column_widgets")
        self.content_view = kwargs.pop("content_view", None)

        col_widget = Text('./button[contains(@class, "pf-c-button")]')
        if column_widgets and 0 not in column_widgets:
            # Do not override column 0 if the user defined it during init
            kwargs["column_widgets"][0] = col_widget
        elif not column_widgets:
            kwargs["column_widgets"] = {0: col_widget}

        super().__init__(*args, **kwargs)

    def _create_row(self, parent, index, logger=None):
        return self.Row(parent, index, self.content_view, logger)


class ExpandableTable(BaseExpandableTable, PatternflyTable):
    pass


class ExpandableColumn(TableColumn):
    EXPAND_LOCATOR = "./button"

    def __init__(
        self,
        parent,
        position,
        content_view=None,
        absolute_position=None,
        logger=None,
        use_index_as_parent=False,
    ):

        super(ExpandableColumn, self).__init__(parent, position, absolute_position, logger=logger)

        # Sometimes expandable content's 'tr' index is dictated by the position of the column
        # Most of the time it is the second index.
        if use_index_as_parent:
            content_parent = Text(parent=self.row, locator=f"./tr[{position + 1}]")
        else:
            content_parent = Text(parent=self.row, locator="./tr[2]")

        if isinstance(content_view, dict):
            content_view = content_view.get(position)

        if content_view:
            self.content = resolve_table_widget(content_parent, content_view)
        else:
            self.content = content_parent

    def __locator__(self):
        """Override the locator to look inside the first 'tr' within the tbody"""
        if self.position == 0:
            # we assume the th column is in the first position
            return "./tr[1]/th[{}]".format(self.position + 1)
        else:
            return "./tr[1]/td[{}]".format(self.position)

    @property
    def is_expandable(self):
        """To check to see if this column is expandable, check to see if a button is in it."""
        try:
            self.browser.element(self.EXPAND_LOCATOR)
        except NoSuchElementException:
            return False
        return True

    def _check_expandable(self):
        if not self.is_expandable:
            raise ColumnNotExpandable(self)

    @property
    def widget(self):
        # TODO: make this configurable?
        if self.is_expandable:
            return Text(locator=self.EXPAND_LOCATOR, parent=self)
        else:
            return None

    @property
    def is_expanded(self):
        self._check_expandable()
        return self.browser.get_attribute("aria-expanded", self.widget).lower()[0] in [
            "1",
            "t",
            "y",
        ]

    def expand(self):
        """Expands the table column."""
        self._check_expandable()
        if not self.is_expanded:
            self.widget.click()
            self.content.wait_displayed()

    def collapse(self):
        """Collapses the table column."""
        self._check_expandable()
        if self.is_expanded:
            self.widget.click()


class CompoundExpandableRow(PatternflyTableRow):
    TABLE_COLUMN_CLS = ExpandableColumn
    HEADER_IN_ROW = "./tr[1]/th[1]"
    # these are the columns that allow for expansion
    EXPANDABLE_COLUMNS = "./tr[1]/td[contains(@class, 'compound-expansion')]"
    Column = ExpandableColumn

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            index = self.table.header_index_mapping[self.table.ensure_normal(item)]
        elif isinstance(item, int):
            index = item
        else:
            raise TypeError("row[] accepts only integers and strings")

        # Typically for this widget the layout is: <th>, <td>, <td>, <td>, and so on...
        if self.has_row_header:
            return self.Column(
                self,
                index,
                self.table.content_view,
                logger=create_item_logger(self.logger, item),
                use_index_as_parent=self.table.use_index_as_parent,
            )

        return super(PatternflyTableRow, self).__getitem__(index)


class CompoundExpandableTable(PatternflyTable):
    """
    This widget is similar to the ExpandableTable, only that it can have any number of
    expandable rows, and the expandable column is not necessarily in the first position.

    https://www.patternfly.org/v4/documentation/react/components/table#
    """

    ROWS = "./tbody"
    ROW_RESOLVER_PATH = "/table/tbody"
    ROW_AT_INDEX = "./tbody[{0}]"
    COLUMN_RESOLVER_PATH = "/tr[0]/td"
    COLUMN_AT_POSITION = "./tr[1]/td[{0}]"
    ROW_TAG = "tbody"
    HEADERS = "./thead/tr/th|./thead/tr/td"

    Row = CompoundExpandableRow

    def __init__(self, *args, **kwargs):
        """Extend init of Table

        Automatically add the 'expand' button widget as column 0.

        Provide additional kwarg for 'content_view', which is used to pass in a WidgetDescriptor
        to be used as the Widget/View for the expanded content of each column.

        Provide additional kwarg for 'use_index_as_parent'. This is necessary because some
        compound expandable tables will put each expandable column's content in a <tr> whose
        index corresponds to the index of the column. Most of the time, the expandable column's
        content is placed in the 2nd <tr>. If not, set this to True when defining the widget.

        Most of the time, the tables that display when a column is expanded are a simple
        PatternflyTable. This means that the same view can be used for multiple columns. E.g.

        .. code-block :: python

            class MyView(View):
                # View with a compound expandable table on it

                class NestedTable(View):
                    table = PatternflyTable()

                table = CompoundExpandableTable(
                    locator=".//table[@aria-label='expandable-table']",
                    content_view=NestedTable(),
                )

        Here, though the table may have many expandable columns, we can just use the same
        NestedTable() to access their content.

        If there need to be different widgets for each expandable column, one can pass a
        dictionary of views where the key is the index of the column to 'content_view'. E.g.

        .. code-block :: python

            class MyView(View):
                # View with a compound expandable table on it

                class NestedTableOne(View):
                    # table that shows after expanding column one
                    table = PatternflyTableOUIA("column-one-table")

                class NestedTableTwo(View):
                    # table that shows after expanding column two
                    table = PatternflyTableOUIA("column-two-table")

                table = CompoundExpandableTable(
                    locator=".//table[@aria-label='expandable-table']",
                    content_view={0: NestedTableOne(), 1: NestedTableTwo()},
                    use_index_as_parent=True
                )

        Here columns one and two have different locators for the table widgets.
        """
        self.content_view = kwargs.pop("content_view", None)
        self.use_index_as_parent = kwargs.pop("use_index_as_parent", False)
        super(CompoundExpandableTable, self).__init__(*args, **kwargs)

    def _create_column(self, parent, position, absolute_position=None, logger=None):
        """Override this if you wish to change column behavior in a child class."""

        if isinstance(self.content_view, dict):
            content_view = self.content_view.get(position)
        else:
            content_view = self.content_view

        return self.Row.Column(
            parent, position, content_view, absolute_position, logger, self.use_index_as_parent
        )



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import SelectItemNotFound
from widgetastic_patternfly4.ouia import ContextSelector

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/contextselector"  # noqa


@pytest.fixture
def view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-contextselector-ouia']"
        contextselector = ContextSelector("Basic")

    return TestView(browser)


def test_contextselector_is_displayed(view):
    assert view.contextselector.is_displayed


def test_contextselector_items(view):
    assert set(view.contextselector.items) == {
        "My Project",
        "OpenShift Cluster",
        "Production Ansible",
        "AWS",
        "Azure",
        "My Project 2",
        "Production Ansible 2",
        "AWS 2",
        "Azure 2",
    }
    assert view.contextselector.has_item("AWS")
    assert not view.contextselector.has_item("Non existing item")


def test_contextselector_open(view):
    assert not view.contextselector.is_open
    view.contextselector.open()
    assert view.contextselector.is_open
    view.contextselector.close()
    assert not view.contextselector.is_open


def test_contextselector_item_select(view):
    view.contextselector.fill("AWS")
    assert view.contextselector.read() == "AWS"
    assert not view.contextselector.is_open
    with pytest.raises(SelectItemNotFound):
        view.contextselector.fill("Non existing item")
    assert not view.contextselector.is_open



from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Checkbox
from widgetastic.widget import GenericLocatorWidget
from widgetastic.widget import ParametrizedView
from widgetastic.widget import View


class BaseCard:
    pass


class Card(BaseCard, GenericLocatorWidget):
    def __init__(self, parent, locator=None, logger=None):
        locator = locator or ".//article[contains(@class, 'pf-c-card')]"
        super().__init__(parent, locator, logger=logger)

    ROOT = ParametrizedLocator("{@locator}")


class CardForCardGroup(ParametrizedView):
    def __init__(self, parent, locator=None, logger=None, **kwargs):
        View.__init__(self, parent, logger=logger, **kwargs)
        self.locator = locator or ".//article[contains(@class, 'pf-c-card')]"

    PARAMETERS = ("position",)

    ROOT = ParametrizedLocator("{@locator}[{position}]")

    def __locator__(self):
        return self.ROOT

    @classmethod
    def all(cls, browser):
        # todo: OUIA versions should return component ids
        elements = browser.elements("./article")
        result = []
        for index, item in enumerate(elements):
            result.append((index + 1,))
        return result


class CardGroup(GenericLocatorWidget, View):
    def __init__(self, parent, locator=None, logger=None, **kwargs):
        View.__init__(self, parent, logger=logger, **kwargs)
        self.locator = locator or './/section[@class="pf-c-page__main-section"]/div'

    cards = ParametrizedView.nested(CardForCardGroup)

    def __iter__(self):
        return iter(self.cards)


class CardCheckBox(Checkbox):
    ROOT = ".//input[@class='pf-c-check__input']"



from .dropdown import Dropdown


class BaseOptionsMenu:
    BUTTON_LOCATOR = (
        ".//button[contains(@class, 'pf-c-options-menu__toggle') or "
        "contains(@class, 'pf-c-options-menu__toggle-button')]"
    )
    ITEMS_LOCATOR = ".//ul[contains(@class, 'pf-c-options-menu__menu')]/li"
    ITEM_LOCATOR = (
        ".//*[contains(@class, 'pf-c-options-menu__menu-item') and normalize-space(.)={}]"
    )
    TEXT_LOCATOR = (
        './/div[contains(@class, "pf-c-options-menu") and '
        'descendant::span[contains(@class, "pf-c-options-menu__toggle-text") and '
        "normalize-space(.)={}]]"
    )
    DEFAULT_LOCATOR = './/div[contains(@class, "pf-c-options-menu")][1]'

    SELECTED_ITEMS_LOCATOR = (
        f"{ITEMS_LOCATOR}/button[.//*[name()='svg'] or descendant::i[not(@hidden)]]"
    )

    @property
    def selected_items(self):
        """Returns a list of all selected items in the options menu."""
        with self.opened():
            return [
                self.browser.text(el) for el in self.browser.elements(self.SELECTED_ITEMS_LOCATOR)
            ]


class OptionsMenu(BaseOptionsMenu, Dropdown):
    pass



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import SelectItemNotFound
from widgetastic_patternfly4.ouia import Select

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/select"


@pytest.fixture
def select(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-select-ouia']"
        select = Select("Single")

    return TestView(browser).select


def test_select_is_displayed(select):
    assert select.is_displayed


@pytest.mark.xfail
def test_select_items(select):
    assert set(select.items) == {"Choose...", "Mr", "Miss", "Mrs", "Ms", "Dr", "Other"}
    assert select.has_item("Mr")
    assert not select.has_item("Non existing item")
    assert select.item_enabled("Miss")


def test_select_open(select):
    assert not select.is_open
    select.open()
    assert select.is_open
    select.close()
    assert not select.is_open


def test_select_item_select(select):
    select.fill("Mr")
    assert select.read() == "Mr"
    assert not select.is_open
    with pytest.raises(SelectItemNotFound):
        select.fill("Non existing item")
    assert not select.is_open



from widgetastic.exceptions import NoSuchElementException
from widgetastic.widget import GenericLocatorWidget


class SwitchDisabled(Exception):
    pass


class BaseSwitch:
    """Represents the Patternfly Switch.

    https://www.patternfly.org/v4/documentation/react/components/switch
    """

    CHECKBOX_LOCATOR = "./input"
    LABEL_ON = "./span[contains(@class, 'pf-m-on')]"
    LABEL_OFF = "./span[contains(@class, 'pf-m-off')]"

    @property
    def selected(self):
        """Returns a boolean detailing if the Switch is on (True) of off (False)."""
        return self.browser.get_attribute("checked", self.CHECKBOX_LOCATOR) is not None

    def _read_locator(self, locator):
        try:
            return self.browser.text(locator)
        except NoSuchElementException:
            return None

    @property
    def label(self):
        """Returns the label of the Switch."""
        if self.selected:
            return self._read_locator(self.LABEL_ON)
        else:
            return self._read_locator(self.LABEL_OFF)

    @property
    def is_enabled(self):
        """Returns a boolean detailing if the switch is enabled."""
        return self.browser.get_attribute("disabled", self.CHECKBOX_LOCATOR) is None

    def fill(self, value):
        """Fills a Switch with the supplied value."""
        if not self.is_enabled:
            raise SwitchDisabled("{} is disabled".format(repr(self)))
        if bool(value) == self.selected:
            return False
        else:
            self.browser.click(self.CHECKBOX_LOCATOR)
            return True

    def read(self):
        """Returns a boolean detailing if the Switch is on (True) of off (False)."""
        return self.selected

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.locator)


class Switch(BaseSwitch, GenericLocatorWidget):
    pass



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Progress

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/progress"
PROGRESS_STATUS_TYPES_WITH_CURRENT_PROGRESS = {
    "success": "100",
    "danger": "33",
    "warning": "90",
    "info": "33",
}


@pytest.fixture(params=PROGRESS_STATUS_TYPES_WITH_CURRENT_PROGRESS.keys())
def progress(browser, request):
    class TestView(View):
        ROOT = ".//main"
        if request.param == "info":
            progress = Progress(locator="(.//div[@class='pf-c-progress'])[1]")
        else:
            progress = Progress(locator=f"(.//div[@class='pf-c-progress pf-m-{request.param}'])[1]")

    return TestView(browser).progress


def test_progress_is_displayed(progress):
    assert progress.is_displayed


def test_progress_status_with_current_progress(progress):
    progress_status_type = progress.status if progress.status != "error" else "danger"
    assert progress_status_type in PROGRESS_STATUS_TYPES_WITH_CURRENT_PROGRESS.keys()
    assert (
        PROGRESS_STATUS_TYPES_WITH_CURRENT_PROGRESS[progress_status_type]
        == progress.current_progress
    )


def test_progress_description(progress):
    assert progress.description == "Title"



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import Switch

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/switch"


@pytest.fixture
def view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-switch-ouia']"
        switch = Switch("Simple Switch")

    return TestView(browser)


def test_switch_is_displayed(view):
    assert view.switch.is_displayed


def test_switch_is_enabled(view):
    assert view.switch.is_enabled


def test_switch_label(view):
    assert view.switch.label == "Message when on"


def test_switch_selected(view):
    assert view.switch.read()


def test_switch_fill(view):
    assert view.switch.selected
    assert view.switch.label == "Message when on"
    assert not view.switch.fill(True)
    assert view.switch.selected
    assert view.switch.label == "Message when on"
    assert view.switch.fill(False)
    assert not view.switch.selected
    assert view.switch.label == "Message when off"
    assert view.switch.fill(True)
    assert view.switch.selected
    assert view.switch.label == "Message when on"



from widgetastic.exceptions import WidgetOperationFailed
from widgetastic.widget import Widget


class BaseBreadCrumb:
    """Represents the Patternfly BreadCrumb.

    https://www.patternfly.org/v4/documentation/react/components/breadcrumb
    """

    ELEMENTS = ".//li"

    @property
    def _path_elements(self):
        return self.browser.elements(self.ELEMENTS)

    @property
    def locations(self):
        """Returns a list of strings of the current location according to the breadcrumbs."""
        return [self.browser.text(loc) for loc in self._path_elements]

    @property
    def active_location(self):
        """Returns the last location in the breadcrumb, or None if there no breadcrumbs."""
        return self.locations[-1] if self.locations else None

    def click_location(self, name, partial=False, handle_alert=False):
        """Clicks a location present in the breadcrumbs by string name.

        Args:
            name: location name
            partial(bool): Whether to use partial match
            handle_alert: handle the browser alert and ensure page safe
        """
        try:
            location = next(
                loc
                for loc in self._path_elements
                if (not partial and (self.browser.text(loc) == name))
                or (partial and (name in self.browser.text(loc)))
            )
        except StopIteration:
            partial_msg = " with partial match" if partial else ""
            locs = f" within locations: {self.locations}"
            raise WidgetOperationFailed(
                f'Breadcrumb location "{name}" not found{partial_msg}{locs}'
            )
        self.browser.click(location, ignore_ajax=handle_alert)
        if handle_alert:
            self.browser.handle_alert(wait=2.0)
            self.browser.plugin.ensure_page_safe()

    def read(self):
        """Return the active location of the breadcrumb."""
        return self.active_location


class BreadCrumb(BaseBreadCrumb, Widget):
    ROOT = './/nav[contains(@class, "pf-c-breadcrumb")]/ol'



from widgetastic.widget import Text
from widgetastic.widget import View

from widgetastic_patternfly4 import Tab

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/tabs"


class TabsTestView(View):
    @View.nested
    class primary(View):
        ROOT = ".//div[@id='ws-react-c-tabs-default']"

        @View.nested
        class tab1(Tab):
            TAB_NAME = "Users"
            content = Text(".")

        @View.nested
        class tab2(Tab):
            TAB_NAME = "Containers"
            content = Text(".")

    @View.nested
    class secondary(View):
        ROOT = ".//div[@id='ws-react-c-tabs-tabs-with-sub-tabs']"

        @View.nested
        class tab1(Tab):
            TAB_NAME = "Users"

            @View.nested
            class secondary1(Tab):
                TAB_NAME = "Secondary tab item 1"
                content = Text(".")

            @View.nested
            class secondary2(Tab):
                TAB_NAME = "Secondary tab item 2"
                content = Text(".")

    @View.nested
    class separate(View):
        ROOT = ".//div[@id='ws-react-c-tabs-separate-content']"

        @View.nested
        class tab1(Tab):
            TAB_NAME = "Tab item 1"
            content = Text(".")

        @View.nested
        class tab2(Tab):
            TAB_NAME = "Tab item 2"
            content = Text(".")

        @View.nested
        class tab3(Tab):
            TAB_NAME = "Tab item 3"
            content = Text(".")


def test_primary_tabs(browser):
    view = TabsTestView(browser)
    assert view.primary.tab1.is_displayed
    # selecting the tab just in case, as other tests could have selected smth different
    view.primary.tab1.select()
    assert view.primary.tab2.is_displayed
    assert view.primary.tab1.is_active()
    assert not view.primary.tab2.is_active()
    assert view.primary.tab1.content.text == "Users"
    view.primary.tab2.select()
    assert not view.primary.tab1.is_active()
    assert view.primary.tab2.is_active()
    assert view.primary.tab2.content.text == "Containers"


def test_secondary_tabs(browser):
    view = TabsTestView(browser)
    assert view.secondary.tab1.is_displayed
    # selecting the tab just in case, as other tests could have selected smth different
    view.primary.tab1.select()
    assert view.secondary.tab1.is_active()
    assert view.secondary.tab1.secondary1.is_displayed
    assert view.secondary.tab1.secondary2.is_displayed
    assert not view.secondary.tab1.secondary1.is_active()
    assert not view.secondary.tab1.secondary2.is_active()
    assert view.secondary.tab1.secondary1.content.text == "Secondary tab item 1 item section"
    view.secondary.tab1.secondary2.select()
    assert not view.secondary.tab1.secondary1.is_active()
    assert view.secondary.tab1.secondary2.is_active()
    assert view.secondary.tab1.secondary2.content.text == "Secondary tab item 2 section"


def test_auto_selected(browser):
    view = TabsTestView(browser)
    # selecting the tab just in case, as other tests could have selected smth different
    view.primary.tab1.select()
    assert not view.primary.tab2.is_active()
    view.primary.tab2.content.read()
    assert view.primary.tab2.is_active()


def test_separate_content_tabs(browser):
    view = TabsTestView(browser)
    view.separate.tab1.select()
    assert view.separate.tab2.is_displayed
    assert view.separate.tab1.is_active()
    assert not view.separate.tab2.is_active()
    assert view.separate.tab1.content.text == "Tab 1 section"
    view.separate.tab2.select()
    assert not view.separate.tab1.is_active()
    assert view.separate.tab2.is_active()
    assert view.separate.tab2.content.text == "Tab 2 section"



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Navigation
from widgetastic_patternfly4 import NavSelectionNotFound

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/navigation"

NAVS = [
    (
        ".//div[@id='ws-react-c-navigation-default']",
        ["Link 1", "Link 2", "Link 3", "Link 4"],
        ["Link 1"],
        "default",
    ),
    (
        ".//div[@id='ws-react-c-navigation-expandable']",
        {
            "Link 1": ["Subnav Link 1", "Subnav Link 2", "Subnav Link 3"],
            "Link 2": ["Custom onClick", "Subnav Link 1", "Subnav Link 2", "Subnav Link 3"],
        },
        ["Link 1", "Subnav Link 1"],
        "expandable",
    ),
    (
        ".//div[@id='ws-react-c-navigation-mixed']",
        {
            "Link 1 (not expandable)": None,
            "Link 2 - expandable": ["Link 1", "Link 2", "Link 3"],
            "Link 3 - expandable": ["Link 1", "Link 2", "Link 3"],
        },
        ["Link 1 (not expandable)"],
        "mixed",
    ),
]


@pytest.fixture(params=NAVS, ids=lambda sample: sample[3])
def data(browser, request):
    locator, tree, currently_selected, _ = request.param

    class TestView(View):
        ROOT = locator

        nav = Navigation(locator="./nav")

    return TestView(browser).nav, tree, currently_selected


def test_navigation(browser, data):
    nav, tree, currently_selected = data
    assert nav.currently_selected == currently_selected
    assert nav.nav_item_tree() == tree


def test_navigation_select(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-navigation-mixed']"
        nav = Navigation(locator="./nav")

    nav = TestView(browser).nav
    nav.select("Link 3 - expandable", "Link 2")
    assert nav.currently_selected == ["Link 3 - expandable", "Link 2"]
    nav.select("Link 1 (not expandable)")
    assert nav.currently_selected == ["Link 1 (not expandable)"]

    # try to select some bogus location and make sure proper exception is raised
    with pytest.raises(NavSelectionNotFound):
        nav.select("This location does not exist")
